#-------------------------------------------------------------------------------
#
# Project: EOxServer <http://eoxserver.org>
# Authors: Fabian Schindler <fabian.schindler@eox.at>
#
#-------------------------------------------------------------------------------
# Copyright (C) 2015 EOX IT Services GmbH
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies of this Software or works derived from this Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.
#-------------------------------------------------------------------------------


from itertools import chain
from datetime import datetime

from lxml.etree import CDATA
from lxml.builder import ElementMaker
from django.template.loader import render_to_string
from django.conf import settings

from eoxserver.core.util.xmltools import etree, NameSpace, NameSpaceMap, typemap
from eoxserver.core.util.timetools import isoformat
from eoxserver.resources.coverages import models
from eoxserver.services.opensearch.formats.base import (
    BaseFeedResultFormat, ns_georss, ns_media, ns_owc
)
from eoxserver.services.opensearch.config import (
    DEFAULT_EOXS_OPENSEARCH_SUMMARY_TEMPLATE
)


# namespace declarations
ns_atom = NameSpace("http://www.w3.org/2005/Atom", None)
ns_opensearch = NameSpace("http://a9.com/-/spec/opensearch/1.1/", "opensearch")
ns_gml = NameSpace("http://www.opengis.net/gml", "gml")
ns_dc = NameSpace("http://purl.org/dc/elements/1.1/", "dc")

# namespace map
nsmap = NameSpaceMap(ns_atom, ns_opensearch, ns_dc, ns_georss, ns_media, ns_owc)

# Element factories
ATOM = ElementMaker(namespace=ns_atom.uri, nsmap=nsmap, typemap=typemap)
OS = ElementMaker(namespace=ns_opensearch.uri, nsmap=nsmap)
GML = ElementMaker(namespace=ns_gml.uri, nsmap=nsmap)
DC = ElementMaker(namespace=ns_dc.uri, nsmap=nsmap)


class AtomResultFormat(BaseFeedResultFormat):
    """ Atom result format.
    """

    mimetype = "application/atom+xml"
    name = "atom"

    def encode(self, request, collection_id, queryset, search_context):

        namespaces = dict(nsmap)
        namespaces.update(search_context.namespaces)
        ATOM = ElementMaker(namespace=ns_atom.uri, nsmap=namespaces)

        tree = ATOM("feed",
            ATOM("id", request.build_absolute_uri()),
            ATOM("title", "%s Search" % collection_id),
            ATOM("description"),
            OS("totalResults", str(search_context.total_count)),
            OS("startIndex", str(search_context.start_index or 0)),
            OS("itemsPerPage", str(search_context.count)),
            OS("Query", role="request", **dict(
                ("{%s}%s" % (namespaces[prefix], name), value)
                for prefix, params in search_context.parameters.items()
                for name, value in params.items()
            )),
            *chain(
                self.encode_feed_links(request, search_context), [
                    self.encode_entry(request, collection_id, item)
                    for item in queryset
                ]
            )
        )
        return etree.tostring(tree, pretty_print=True)

    def encode_entry(self, request, collection_id, item):
        entry = ATOM("entry",
            ATOM("title", item.identifier),
            ATOM("id", self._create_self_link(request, collection_id, item)),
            DC("identifier", item.identifier),
            *self.encode_spatio_temporal(item)
        )
        entry.extend(self.encode_item_links(request, collection_id, item))
        entry.append(self.encode_summary(request, collection_id, item))
        return entry

    def encode_summary(self, request, collection_id, item):
        template_name = getattr(
            settings, 'EOXS_OPENSEARCH_SUMMARY_TEMPLATE',
            DEFAULT_EOXS_OPENSEARCH_SUMMARY_TEMPLATE
        )

        metadata = []
        coverages = []

        if isinstance(item, models.Coverage):
            coverages = [item]
        elif isinstance(item, models.Product):
            coverages = item.coverages.all()
            metadata = [
                (
                    name.replace('_', ' ').title(),
                    isoformat(value) if isinstance(value, datetime) else str(value)
                )
                for name, value in models.product_get_metadata(item)
            ]

        eo_om_item = item.metadata_items.filter(
            format__in=['eogml', 'eoom', 'text/xml'],
            semantic__isnull=False
        ).first()
        if eo_om_item is not None:
            eo_om_link = self._make_metadata_href(request, item, eo_om_item)
        else:
            eo_om_link = None

        template_params = {
            'item': item,
            'metadata': metadata,
            'atom': self._create_self_link(request, collection_id, item),
            'wms_capabilities': self._create_wms_capabilities_link(request, item),
            'map_small': self._create_map_link(request, item, 100),
            'map_large': self._create_map_link(request, item, 500),
            'eocoveragesetdescription': self._create_eo_coverage_set_description(
                request, item
            ),
            'coverages': [{
                'identifier': coverage.identifier,
                'description': self._create_coverage_description_link(
                    request, coverage
                ),
                'coverage': self._create_coverage_link(
                    request, coverage
                )}
                for coverage in coverages
            ],
            'download_link': self._create_download_link(
                request, item
            ) if isinstance(item, models.Product) else None,
            'eo_om_link': eo_om_link,
        }

        return ATOM("summary",
            CDATA(
                render_to_string(
                    template_name, template_params,
                    request=request
                )
            ),
            type="html"
        )



#
#  Various helper subroutines needed by execute handler.
#
# Project: EOxServer <http://eoxserver.org>
# Authors: Martin Paces <martin.paces@eox.at>
#
#-------------------------------------------------------------------------------
# Copyright (C) 2016 EOX IT Services GmbH
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies of this Software or works derived from this Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.
#-------------------------------------------------------------------------------

from django.utils.six import iteritems, itervalues
from eoxserver.services.ows.wps.util import OrderedDict
from eoxserver.services.ows.wps.parameters import (
    fix_parameter, LiteralData, BoundingBoxData, ComplexData,
    InputReference, InputData, RequestParameter,
)
from eoxserver.services.ows.wps.exceptions import (
    InvalidInputError, MissingRequiredInputError,
    InvalidInputReferenceError, InvalidInputValueError,
    InvalidOutputError, InvalidOutputDefError,
)


def parse_params(param_defs):
    """ Parse process's inputs/outputs parameter definitions. """
    if isinstance(param_defs, dict):
        param_defs = iteritems(param_defs)
    return OrderedDict(
        (param.identifier, (name, param)) for name, param in (
            (name, fix_parameter(name, param)) for name, param in param_defs
        )
    )


def check_invalid_inputs(inputs, input_defs):
    """Check that there are no undefined inputs."""
    invalids = set(inputs) - set(input_defs)
    if invalids:
        raise InvalidInputError(invalids.pop())
    return inputs


def check_invalid_outputs(outputs, output_defs):
    """Check that there are no undefined outputs."""
    invalids = set(outputs) - set(output_defs)
    if invalids:
        raise InvalidOutputError(invalids.pop())
    return outputs


def resolve_request_parameters(inputs, input_defs, request):
    """ Resolve request parameters. """
    for identifier, (_, input_def) in iteritems(input_defs):
        if isinstance(input_def, RequestParameter):
            inputs[identifier] = input_def.parse_request(request)
    return inputs


def decode_raw_inputs(raw_inputs, input_defs, resolver):
    """ Iterates over all input options stated in the process and parses
        all given inputs. This also includes resolving of references.
    """
    decoded_inputs = {}
    for identifier, (name, input_def) in iteritems(input_defs):
        raw_value = raw_inputs.get(identifier)
        if isinstance(input_def, RequestParameter):
            value = raw_value
        elif raw_value is not None:
            if isinstance(raw_value, InputReference):
                if not input_def.resolve_input_references:
                    # do not resolve reference if this is explicitly required
                    # by the input definition
                    decoded_inputs[name] = raw_value
                    continue
                try:
                    raw_value = _resolve_reference(raw_value, resolver)
                except ValueError as exc:
                    raise InvalidInputReferenceError(identifier, exc)
            try:
                value = _decode_input(input_def, raw_value)
            except ValueError as exc:
                raise InvalidInputValueError(identifier, exc)
        elif input_def.is_optional:
            value = getattr(input_def, 'default', None)
        else:
            raise MissingRequiredInputError(identifier)
        decoded_inputs[name] = value
    return decoded_inputs


def _decode_input(input_def, raw_value):
    """ Decode raw input and check it against the allowed values."""
    if isinstance(input_def, LiteralData):
        return input_def.parse(raw_value.data, raw_value.uom)
    elif isinstance(input_def, BoundingBoxData):
        return input_def.parse(raw_value.data)
    elif isinstance(input_def, ComplexData):
        return input_def.parse(
            raw_value.data, raw_value.mime_type,
            raw_value.schema, raw_value.encoding, urlsafe=raw_value.asurl
        )
    else:
        raise TypeError("Unsupported parameter type %s!"%(type(input_def)))


def _resolve_reference(input_ref, resolver):
    """ Get the input passed as a reference."""
    # prepare HTTP/POST request
    if input_ref.method == "POST":
        if input_ref.body_href is not None:
            input_ref.body = resolver(
                input_ref.body_href, None, input_ref.headers
            )
        if input_ref.body is not None:
            ValueError("Missing the POST request body!")
    else:
        input_ref.body = None
    data = resolver(input_ref.href, input_ref.body, input_ref.headers)
    return InputData(
        data=data,
        identifier=input_ref.identifier,
        title=input_ref.title,
        abstract=input_ref.abstract,
        mime_type=input_ref.mime_type,
        encoding=input_ref.encoding,
        schema=input_ref.schema,
    )


def decode_output_requests(response_form, output_defs):
    """ Complex data format selection (mimeType, encoding, schema)
        is passed as an input to the process
    """
    output_requests = {}
    for identifier, (name, output_def) in iteritems(output_defs):
        request = response_form.get_output(identifier)
        if isinstance(output_def, ComplexData):
            format_ = output_def.get_format(
                request.mime_type,
                request.encoding,
                request.schema
            )
            if format_ is None:
                raise InvalidOutputDefError(identifier, (
                    "Invalid complex data format! "
                    "mimeType=%r encoding=%r schema=%r" % (
                        request.mime_type,
                        request.encoding,
                        request.schema
                    )
                ))
            output_requests[name] = {
                "mime_type": format_.mime_type,
                "encoding": format_.encoding,
                "schema": format_.schema,
                "as_reference": request.as_reference,
            }
    return output_requests


def pack_outputs(outputs, response_form, output_defs):
    """ Collect data, output declaration and output request for each item."""
    # Normalize the outputs to a dictionary.
    if not isinstance(outputs, dict):
        if len(output_defs) == 1:
            outputs = {list(itervalues(output_defs))[0][0]: outputs}
        else:
            outputs = dict(outputs)
    # Pack the outputs to a tuple containing:
    #   - the output data (before encoding)
    #   - the process output declaration (ProcessDescription/Output)
    #   - the output's requested form (RequestForm/Output)
    packed = OrderedDict()
    for identifier, (name, output_def) in iteritems(output_defs):
        request = response_form.get_output(identifier)
        result = outputs.get(name)
        # TODO: Can we silently skip the missing outputs? Check the standard!
        if result is not None:
            packed[identifier] = (result, output_def, request)
        elif isinstance(output_def, LiteralData) and output_def.default is not None:
            packed[identifier] = (output_def.default, output_def, request)
    return packed




# -----------------------------------------------------------------------------
#
# Project: EOxServer <http://eoxserver.org>
# Authors: Martin Paces <martin.paces@eox.at>
#
# -----------------------------------------------------------------------------
# Copyright (C) 2014 EOX IT Services GmbH
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies of this Software or works derived from this Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.
# -----------------------------------------------------------------------------

import csv
from datetime import datetime

from django.db.models import Q
from django.contrib.gis.db.models.functions import Envelope

from eoxserver.core import Component, implements
from eoxserver.core.util.timetools import isoformat

from eoxserver.services.ows.wps.interfaces import ProcessInterface
from eoxserver.services.ows.wps.parameters import (
    LiteralData, ComplexData, CDAsciiTextBuffer, FormatText
)
from eoxserver.services.ows.wps.exceptions import InvalidInputValueError

from eoxserver.resources.coverages import models


class GetTimeDataProcess(Component):
    """ GetTimeDataProcess defines a WPS process needed by the EOxClient
        time-slider componenet """

    implements(ProcessInterface)

    identifier = "getTimeData"
    title = "Get times of collection coverages."
    decription = ("Query collection and get list of coverages and their times "
                  "and spatial extents. The process is used by the "
                  "time-slider of the EOxClient (web client).")

    metadata = {}
    profiles = ['EOxServer:GetTimeData']

    inputs = {
        "collection": LiteralData(
            "collection",
            title="Collection name (a.k.a. dataset-series identifier)."),
        "begin_time": LiteralData(
            "begin_time",
            datetime,
            optional=True,
            title="Optional start of the time interval."
        ),
        "end_time": LiteralData(
            "end_time",
            datetime,
            optional=True,
            title="Optional end of the time interval."
        ),
    }

    outputs = {
        "times": ComplexData(
            "times",
            formats=(FormatText('text/csv'), FormatText('text/plain')),
            title=(
                "Comma separated list of collection's coverages, "
                "their extents and times."
            ),
            abstract=(
                "NOTE: The use of the 'text/plain' format is "
                "deprecated! This format will be removed!'"
            )
        )
    }

    @staticmethod
    def execute(collection, begin_time, end_time, **kwarg):
        """ The main execution function for the process.
        """

        # get the dataset series matching the requested ID
        try:
            model = models.EOObject.objects.filter(
                identifier=collection
            ).select_subclasses().get()
        except models.EOObject.DoesNotExist:
            raise InvalidInputValueError(
                "collection", "Invalid collection name '%s'!" % collection
            )

        if isinstance(model, (models.Collection, models.Mosaic)):
            if isinstance(model, models.Collection):
                qs = models.EOObject.objects.filter(
                    Q(  # coverages
                        coverage__isnull=False,
                        coverage__collections=model,
                    ) | Q (  # products
                        product__isnull=False,
                        product__collections=model,
                    )
                )
            else:
                qs = model.coverages

            # prepare coverage query set
            if end_time is not None:
                qs = qs.filter(begin_time__lte=end_time)
            if begin_time is not None:
                qs = qs.filter(end_time__gte=begin_time)
            qs = qs.order_by(
                'begin_time', 'end_time'
            ).annotate(
                envelope=Envelope('footprint')
            ).values_list(
                "begin_time", "end_time", "identifier", "envelope"
            )

        else:
            qs = [(
                model.begin_time,
                model.end_time,
                model.identifier,
                model.footprint
            )]

        # create the output
        output = CDAsciiTextBuffer()
        writer = csv.writer(output, quoting=csv.QUOTE_ALL)
        header = ["starttime", "endtime", "bbox", "identifier"]
        writer.writerow(header)

        for starttime, endtime, identifier, bbox in qs:
            writer.writerow([
                isoformat(starttime),
                isoformat(endtime),
                bbox.extent if not bbox.empty else '()',
                identifier
            ])

        return output



# ------------------------------------------------------------------------------
#
# Project: EOxServer <http://eoxserver.org>
# Authors: Fabian Schindler <fabian.schindler@eox.at>
#
# ------------------------------------------------------------------------------
# Copyright (C) 2019 EOX IT Services GmbH
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies of this Software or works derived from this Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.
# ------------------------------------------------------------------------------

import json

from django.core.management.base import CommandError, BaseCommand
from django.db import transaction

from eoxserver.backends import models as backends
from eoxserver.backends.storage_auths import get_handler_for_model
from eoxserver.resources.coverages.management.commands import (
    CommandOutputMixIn, SubParserMixIn
)


class Command(CommandOutputMixIn, SubParserMixIn, BaseCommand):
    """ Command to manage storages authorizations.
        This command uses sub-commands for the
        specific tasks: create, delete
    """
    def add_arguments(self, parser):
        create_parser = self.add_subparser(parser, 'create')
        delete_parser = self.add_subparser(parser, 'delete')

        # name is a common argument
        for parser in [create_parser, delete_parser]:
            parser.add_argument(
                'name', nargs=1, help='The storage auth name'
            )

        create_parser.add_argument(
            'url', nargs=1,
            help='The storage auth location in a URL format. Mandatory.'
        )
        create_parser.add_argument(
            '--type', '-t', dest='type_name', default=None,
            help=(
                'The storage auth type. Optional. Default is auto-detect the '
                'type.'
            )
        )
        create_parser.add_argument(
            '--parameter', '-p', dest='parameters', default=None, nargs='+',
            action='append',
            help='Pass arguments for that specific storage auth item.',
        )

        create_parser.add_argument(
            '--check', dest='check', default=False, action='store_true',
            help='Check access to the storage auth.',
        )

    @transaction.atomic
    def handle(self, subcommand, name, *args, **kwargs):
        """ Dispatch sub-commands: create, delete, insert, exclude, purge.
        """
        name = name[0]
        if subcommand == "create":
            self.handle_create(name, *args, **kwargs)
        elif subcommand == "delete":
            self.handle_delete(name, *args, **kwargs)

    def handle_create(self, name, url, type_name, parameters, check, **kwargs):
        """ Handle the creation of a new storage.
        """
        url = url[0]

        def parse_parameter(key, value=None, *extra):
            if extra:
                raise CommandError('Extra arguments for parameter found')

            if value is None:
                key, _, value = key.partition('=')

            return key.replace('-', '_'), value

        parameters = dict(
            parse_parameter(*param)
            for param in parameters
        )

        storage_auth = backends.StorageAuth(
            name=name,
            url=url,
            storage_auth_type=type_name,
            auth_parameters=json.dumps(parameters),
        )
        storage_auth.full_clean()
        storage_auth.save()

        if check:
            _ = get_handler_for_model(storage_auth)
            # TODO perform check

        self.print_msg(
            'Successfully created storage auth %s (%s)' % (name, type_name)
        )

    def handle_delete(self, name, **kwargs):
        """ Handle the deletion of a storage
        """
        try:
            storage_auth = backends.StorageAuth.objects.get(name=name)
        except backends.StorageAuth.DoesNotExist:
            raise CommandError('No such storage with name %r' % name)
        type_name = storage_auth.storage_auth_type
        storage_auth.delete()

        self.print_msg(
            'Successfully deleted storage auth %s (%s)' % (name, type_name)
        )



# ------------------------------------------------------------------------------
#
# Project: EOxServer <http://eoxserver.org>
# Authors: Fabian Schindler <fabian.schindler@eox.at>
#
# ------------------------------------------------------------------------------
# Copyright (C) 2017 EOX IT Services GmbH
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies of this Software or works derived from this Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.
# ------------------------------------------------------------------------------

from eoxserver.core.decoders import kvp
from eoxserver.core.util.xmltools import NameSpace
from eoxserver.services import filters, ecql


class CQLExtension(object):
    """ Implementation of the OpenSearch `'EO' extension
    <http://docs.opengeospatial.org/is/13-026r8/13-026r8.html>`_.
    """

    namespace = NameSpace(
        "http://a9.com/-/opensearch/extensions/cql/1.0/", "cql"
    )

    def filter(self, qs, parameters):
        mapping, mapping_choices = filters.get_field_mapping_for_model(qs.model)
        decoder = CQLExtensionDecoder(parameters)

        cql_text = decoder.cql
        if cql_text:
            ast = ecql.parse(cql_text)
            filter_expressions = ecql.to_filter(ast, mapping, mapping_choices)

            qs = qs.filter(filter_expressions)

        return qs

    def get_schema(self, collection=None, model_class=None):
        return (
            dict(name="cql", type="cql", profiles=[
                dict(
                    href="http://www.opengis.net/csw/3.0/cql",
                    title=(
                        "CQL (Common Query Language) is a query language "
                        "created by the OGC for the Catalogue Web Services "
                        "specification."
                    )
                )
            ]),
        )


class CQLExtensionDecoder(kvp.Decoder):
    cql = kvp.Parameter(num="?", type=str)



#-------------------------------------------------------------------------------
#
# Project: EOxServer <http://eoxserver.org>
# Authors: Fabian Schindler <fabian.schindler@eox.at>
#          Martin Paces <martin.paces@eox.at>
#
#-------------------------------------------------------------------------------
# Copyright (C) 2011-2014 EOX IT Services GmbH
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies of this Software or works derived from this Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.
#-------------------------------------------------------------------------------

from eoxserver.contrib import mapserver as ms
from eoxserver.resources.coverages import models

from eoxserver.services.mapserver.wms.layerfactories.base import (
    BaseCoverageLayerFactory
)


class CoverageMaskedLayerFactory(BaseCoverageLayerFactory):
    handles = (models.RectifiedDataset, models.RectifiedStitchedMosaic)
    suffixes = ("_masked",)
    requires_connection = True

    def generate(self, eo_object, group_layer, suffix, options):
        mask_layer_name = eo_object.identifier + "__mask__"

        # handle the mask layers

        # get the applicable sematics
        mask_semantics = ("polygonmask",)
        mask_items = eo_object.data_items.all()
        for mask_semantic in mask_semantics:
            mask_items = mask_items.filter(semantic__startswith=mask_semantic)

        # layer creating closure
        def _create_mask_polygon_layer(name):
            mask_layer = ms.layerObj()
            mask_layer.name = name
            mask_layer.type = ms.MS_LAYER_POLYGON

            mask_layer.setMetaData("eoxs_geometry_reversed", "true")

            cls = ms.classObj(mask_layer)
            style = ms.styleObj(cls)
            style.color.setRGB(0, 0, 0)
            return mask_layer

        # multiple masks shall be grouped by a group layer
        if len(mask_items) > 1:

            # more than one mask, requires a mask group layer
            mask_layer = ms.Layer(mask_layer_name)

            yield (mask_layer, ())

            # generate mask layers
            for i, mask_item in enumerate(mask_items):

                mask_sublayer = _create_mask_polygon_layer(
                    "%s%2.2d" % (mask_layer_name, i)
                )
                mask_sublayer.group = mask_layer.name

                yield (mask_sublayer, (mask_item,))

        # single mask shall be used directly as a "group" layer
        elif len(mask_items) == 1:

            mask_layer = _create_mask_polygon_layer(mask_layer_name)

            yield (mask_layer, (mask_items[0],))

        # no mask at all
        else:
            mask_layer = None

        # handle the image layers
        super_items = super(CoverageMaskedLayerFactory, self).generate(
            eo_object, group_layer, suffix, options
        )

        for layer, data_items in super_items:

            # if we do have a mask, reference it in the layer
            if mask_layer:
                layer.mask = mask_layer.name

            # fix the layer name by appending the right suffix
            layer.name = layer.name + suffix
            if layer.group:
                layer.group = layer.group + suffix

            # "re-yield" the layer and its items
            yield (layer, data_items)

    def generate_group(self, name):
        layer = ms.layerObj()
        layer.name = name
        layer.type = ms.MS_LAYER_RASTER
        return layer



#-------------------------------------------------------------------------------
#
# Project: EOxServer <http://eoxserver.org>
# Authors: Fabian Schindler <fabian.schindler@eox.at>
#
#-------------------------------------------------------------------------------
# Copyright (C) 2011 EOX IT Services GmbH
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies of this Software or works derived from this Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.
#-------------------------------------------------------------------------------


class WMSCapabilitiesRendererInterface(object):
    """ Interface for WMS compatible capabilities renderers.
    """

    def render(self, collections, coverages, request_values):
        """ Render a capabilities document, containing metadata of the given 
            collections and coverages.
        """


class WMSMapRendererInterface(object):
    """ Interface for WMS compatible map renderers.
    """

    @property
    def suffixes(self):
        """ Return a list of supported layer suffixes for this renderer.
        """

    def render(self, layer_groups, request_values, **options):
        """ Render the given layer hierarchy with the provided request values 
            and further options.

            ``options`` contains relevant options such as specified bands.
        """

class WMSFeatureInfoRendererInterface(object):
    """ Interface for WMS compatible feature info renderers.
    """

    @property
    def suffixes(self):
        """ Return a list of supported layer suffixes for this renderer.
        """

    def render(self, layer_groups, request_values, **options):
        """ Render the given layer hierarchy with the provided request values 
            and further options.

            ``options`` contains relevant options such as specified bands.
        """

class WMSLegendGraphicRendererInterface(object):
    """ Interface for WMS compatible legend graphic renderers.
    """

    @property
    def suffixes(self):
        """ Return a list of supported layer suffixes for this renderer.
        """

    def render(self, collection, eo_object, request_values, **options):
        """ Render the given collection and coverage with the provided request
            values and further options.

            ``options`` contains relevant options such as specified bands.
        """




#-------------------------------------------------------------------------------
#
# WPS 1.0 execute response raw encoder
#
# Project: EOxServer <http://eoxserver.org>
# Authors: Fabian Schindler <fabian.schindler@eox.at>
#          Martin Paces <martin.paces@eox.at>
#
#-------------------------------------------------------------------------------
# Copyright (C) 2013 EOX IT Services GmbH
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies of this Software or works derived from this Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.
#-------------------------------------------------------------------------------

import types
try:
    from StringIO import StringIO
except ImportError:
    from io import StringIO

from eoxserver.services.result import (
    to_http_response, ResultItem,
)
from eoxserver.services.ows.wps.parameters import (
    LiteralData, ComplexData, BoundingBoxData,
)
from eoxserver.services.ows.wps.exceptions import InvalidOutputValueError
from django.utils.six import string_types, itervalues


class WPS10ExecuteResponseRawEncoder(object):
    """ WPS 1.0 raw output Execute response encoder. """

    @staticmethod
    def serialize(result_items, **kwargs):
        """ Serialize the result items to the HTTP response object. """
        # pylint: disable=unused-argument
        return to_http_response(result_items)

    def __init__(self, resp_form):
        self.content_type = None
        self.resp_form = resp_form

    def encode_response(self, results):
        """Pack the raw execute response."""
        outputs = []
        for data, prm, req in itervalues(results):
            if prm.identifier in self.resp_form:
                outputs.append(_encode_raw_output(data, prm, req))

        if len(outputs) == 1:
            self.content_type = outputs[0].content_type
        else:
            self.content_type = "multipart/related"

        return outputs

#-------------------------------------------------------------------------------

class ResultAlt(ResultItem):
    """ Alternative implementation of the result buffer. The object can be
    initialized with a byte-string, sequence or generator of byte-strings,
    or seekable file(-like) object.
    """

    def __init__(self, buf, content_type=None, filename=None, identifier=None,
                 close=False, headers=None):
        # pylint: disable=too-many-arguments
        ResultItem.__init__(self, content_type, filename, identifier)
        if isinstance(buf, string_types) or isinstance(buf, bytes):
            self._file = StringIO(buf.decode('utf-8'))  # make sure a byte string is passed
        elif isinstance(buf, (tuple, list, types.GeneratorType)):
            tmp = StringIO()
            for chunk in buf:
                tmp.write(chunk)
            self._file = tmp
        else:
            self._file = buf
        self._close = close
        self.headers = headers or []

    def __del__(self):
        if self._close:
            self._file.close()

    def __len__(self):
        self._file.seek(0, 2)
        return self._file.tell()

    @property
    def data_file(self):
        self._file.seek(0)
        return self._file

    @property
    def data(self):
        return self.data_file.read()

    def chunked(self, chunksize):
        if chunksize < 0:
            raise ValueError("Invalid chunk-size %d!" % chunksize)
        data_file = self.data_file
        for chunk in iter(lambda: data_file.read(chunksize), ''):
            yield chunk

#-------------------------------------------------------------------------------


def _encode_raw_output(data, prm, req):
    """ Encode a raw output item."""
    if isinstance(prm, LiteralData):
        return _encode_raw_literal(data, prm, req)
    elif isinstance(prm, BoundingBoxData):
        return _encode_raw_bbox(data, prm, req)
    elif isinstance(prm, ComplexData):
        return _encode_raw_complex(data, prm)
    raise TypeError("Invalid output type! %r"%(prm))


def _encode_raw_literal(data, prm, req):
    """ Encode a raw literal."""
    content_type = "text/plain" if req.mime_type is None else req.mime_type
    content_type = "%s; charset=utf-8" % content_type
    try:
        encoded_data = prm.encode(data, req.uom or prm.default_uom, 'utf-8')
    except (ValueError, TypeError) as exc:
        raise InvalidOutputValueError(prm.identifier, exc)
    return ResultAlt(
        encoded_data, identifier=prm.identifier, content_type=content_type
    )


def _encode_raw_bbox(data, prm, req):
    """ Encode a raw bounding box."""
    content_type = "text/plain" if req.mime_type is None else req.mime_type
    content_type = "%s; charset=utf-8"%content_type
    try:
        encoded_data = prm.encode_kvp(data).encode('utf-8')
    except (ValueError, TypeError) as exc:
        raise InvalidOutputValueError(prm.identifier, exc)
    return ResultAlt(
        encoded_data, identifier=prm.identifier,
        content_type="text/plain" if req.mime_type is None else req.mime_type
    )


def _encode_raw_complex(data, prm):
    """ Encode raw complex data."""
    payload, content_type = prm.encode_raw(data)
    return ResultAlt(
        payload, identifier=prm.identifier, content_type=content_type,
        filename=getattr(data, "filename", None),
        headers=getattr(data, "headers", None),
    )



# ------------------------------------------------------------------------------
#
# Project: EOxServer <http://eoxserver.org>
# Authors: Fabian Schindler <fabian.schindler@eox.at>
#          Stephan Meissl <stephan.meissl@eox.at>
#          Stephan Krause <stephan.krause@eox.at>
#
# ------------------------------------------------------------------------------
# Copyright (C) 2011 EOX IT Services GmbH
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies of this Software or works derived from this Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.
# ------------------------------------------------------------------------------

from django import forms
from django.contrib import admin

from eoxserver.backends.storages import get_handlers
from eoxserver.backends import models


# ==============================================================================
# choice helpers
# ==============================================================================

def get_storage_type_choices():
    return [
        (handler.name, handler.name) for handler in get_handlers()
    ]


# ==============================================================================
# Forms
# ==============================================================================


class StorageForm(forms.ModelForm):
    """ Form for `Storages`. Overrides the `format` formfield and adds choices
        dynamically.
    """

    def __init__(self, *args, **kwargs):
        super(StorageForm, self).__init__(*args, **kwargs)
        self.fields['storage_type'] = forms.ChoiceField(
            choices=[("---------", None)] + get_storage_type_choices()
        )


# ==============================================================================
# Admins
# ==============================================================================


class StorageAdmin(admin.ModelAdmin):
    form = StorageForm
    model = models.Storage

    def save_model(self, request, obj, form, change):
        if not obj.name:
            obj.name = None
        super(StorageAdmin, self).save_model(request, obj, form, change)

admin.site.register(models.Storage, StorageAdmin)
admin.site.register(models.StorageAuth)



#-------------------------------------------------------------------------------
#
# Project: EOxServer <http://eoxserver.org>
# Authors: Fabian Schindler <fabian.schindler@eox.at>
#
#-------------------------------------------------------------------------------
# Copyright (C) 2013 EOX IT Services GmbH
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies of this Software or works derived from this Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.
#-------------------------------------------------------------------------------

""" This module contains utilities to easily import hierarchies of packages and
modules.
"""

import logging
import traceback
import pkgutil

try:
    # Python >= 3.1
    from importlib import import_module
except:
    try:
        # Django versions >= 1.9
        from django.utils.module_loading import import_module
    except ImportError:
        # Django versions < 1.9
        from django.utils.importlib import import_module


logger = logging.getLogger(__name__)


def easy_import(module_path):
    """ Utility function to import one or more modules via a given module path.
        The last component of the module path can also be a '*' or a '**'
        character string which imports all submodules of the package either
        recursively (with '**') or not (with '*').

        :param module_path: a typical python module path in the dotted notation.
                            wildcards can be appeded at the last level.
    """

    parts = module_path.split(".")
    if parts[-1] == "*":
        import_modules(".".join(parts[:-1]))
    elif parts[-1] == "**":
        import_recursive(".".join(parts[:-1]))
    else:
        try:
            import_module(module_path)
            logger.debug("Imported module '%s'." % module_path)
        except ImportError:
            logger.error("Failed to import module '%s'." % module_path)
            logger.debug(traceback.format_exc())


def import_modules(base_module_path):
    """ Helper function to import all direct submodules within a package. This
        function is not recursive.

        :param base_module_path: the base module path in the dotted notation.
    """

    path = import_module(base_module_path).__path__
    for loader, module_name, is_pkg in pkgutil.iter_modules(path):
        full_path = "%s.%s" % (base_module_path, module_name)
        try:
            loader.find_module(module_name).load_module(module_name)
            logger.debug("Imported module '%s'." % full_path)
        except ImportError:
            logger.error("Failed to import module '%s'." % full_path)
            logger.debug(traceback.format_exc())


def import_recursive(base_module_path):
    """  Helper function to recursively import all submodules and packages.

        :param base_module_path: the base module path in the dotted notation.
    """

    path = import_module(base_module_path).__path__
    for loader, module_name, is_pkg in pkgutil.walk_packages(path):
        full_path = "%s.%s" % (base_module_path, module_name)
        try:
            loader.find_module(module_name).load_module(module_name)
            logger.debug("Imported module '%s'." % full_path)
        except ImportError:
            logger.error("Failed to import module '%s'." % full_path)
            logger.debug(traceback.format_exc())


def import_string(dotted_path):
    """
    Import a dotted module path and return the attribute/class designated by the
    last name in the path. Raise ImportError if the import failed.
    """
    try:
        module_path, class_name = dotted_path.rsplit('.', 1)
    except ValueError as err:
        raise ImportError("%s doesn't look like a module path" % dotted_path)

    module = import_module(module_path)

    try:
        return getattr(module, class_name)
    except AttributeError as err:
        raise ImportError('Module "%s" does not define a "%s" attribute/class' % (
            module_path, class_name)
        )



# ------------------------------------------------------------------------------
#
# Project: EOxServer <http://eoxserver.org>
# Authors: Fabian Schindler <fabian.schindler@eox.at>
#          Stephan Meissl <stephan.meissl@eox.at>
#          Stephan Krause <stephan.krause@eox.at>
#
# ------------------------------------------------------------------------------
# Copyright (C) 2011 EOX IT Services GmbH
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies of this Software or works derived from this Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.
# ------------------------------------------------------------------------------

# pep8: disable=E501

import json
from datetime import datetime
import re

from django.conf import settings
from django.core.exceptions import ValidationError
from django.core.validators import RegexValidator
from django.contrib.gis.db import models
from django.contrib.gis.db.models import Extent, Union
from django.contrib.gis.geos import Polygon
from django.db.models import Min, Max, Q, F, ExpressionWrapper
from django.db.models.functions import Cast
from django.utils.timezone import now
from django.utils.encoding import python_2_unicode_compatible
from model_utils.managers import InheritanceManager

from eoxserver.backends import models as backends
from eoxserver.core.util.timetools import isoformat
from eoxserver.render.browse.generate import (
    parse_expression, extract_fields, BandExpressionError
)


mandatory = dict(null=False, blank=False)
optional = dict(null=True, blank=True)
searchable = dict(null=True, blank=True, db_index=True)

optional_protected = dict(null=True, blank=True, on_delete=models.PROTECT)
mandatory_protected = dict(null=False, blank=False, on_delete=models.PROTECT)

optional_indexed = dict(blank=True, null=True, db_index=True)
common_value_args = dict(
    on_delete=models.SET_NULL, null=True, blank=True,
    related_name="metadatas"
)

name_validators = [
    RegexValidator(
        re.compile(r'^[a-zA-z_][a-zA-Z0-9_]*$'),
        message="This field must contain a valid Name."
    )
]

identifier_validators = [
    RegexValidator(
        re.compile(r'^[a-zA-z_][a-zA-Z0-9_.-]*$'),
        message="This field must contain a valid NCName."
    )
]


def band_expression_validator(band_expression):
    if not band_expression:
        return

    try:
        parse_expression(band_expression)
    except BandExpressionError as e:
        raise ValidationError(str(e))


# ==============================================================================
# "Type" models
# ==============================================================================


class FieldType(models.Model):
    coverage_type = models.ForeignKey('CoverageType', on_delete=models.CASCADE, related_name='field_types', **mandatory)
    index = models.PositiveSmallIntegerField(**mandatory)
    identifier = models.CharField(max_length=512, validators=identifier_validators, **mandatory)
    description = models.TextField(**optional)
    definition = models.CharField(max_length=512, **optional)
    unit_of_measure = models.CharField(max_length=64, **optional)
    wavelength = models.FloatField(**optional)
    significant_figures = models.PositiveSmallIntegerField(**optional)
    numbits = models.PositiveSmallIntegerField(**optional)
    signed = models.BooleanField(default=True, **mandatory)
    is_float = models.BooleanField(default=False, **mandatory)

    class Meta:
        ordering = ('index',)
        unique_together = (
            ('index', 'coverage_type'), ('identifier', 'coverage_type')
        )

    def __str__(self):
        return self.identifier


class AllowedValueRange(models.Model):
    field_type = models.ForeignKey(FieldType, on_delete=models.CASCADE, related_name='allowed_value_ranges')
    start = models.FloatField(**mandatory)
    end = models.FloatField(**mandatory)


class NilValue(models.Model):
    NIL_VALUE_CHOICES = (
        ("http://www.opengis.net/def/nil/OGC/0/inapplicable", "Inapplicable (There is no value)"),
        ("http://www.opengis.net/def/nil/OGC/0/missing", "Missing"),
        ("http://www.opengis.net/def/nil/OGC/0/template", "Template (The value will be available later)"),
        ("http://www.opengis.net/def/nil/OGC/0/unknown", "Unknown"),
        ("http://www.opengis.net/def/nil/OGC/0/withheld", "Withheld (The value is not divulged)"),
        ("http://www.opengis.net/def/nil/OGC/0/AboveDetectionRange", "Above detection range"),
        ("http://www.opengis.net/def/nil/OGC/0/BelowDetectionRange", "Below detection range")
    )
    field_types = models.ManyToManyField(FieldType, related_name='nil_values', blank=True)
    value = models.CharField(max_length=512, **mandatory)
    reason = models.CharField(max_length=512, choices=NIL_VALUE_CHOICES, **mandatory)


class MaskType(models.Model):
    name = models.CharField(max_length=512, validators=name_validators, **mandatory)
    product_type = models.ForeignKey('ProductType', on_delete=models.CASCADE, related_name='mask_types', **mandatory)
    validity = models.BooleanField(default=False)

    def __str__(self):
        return self.name

    class Meta:
        unique_together = (
            ('name', 'product_type'),
        )


class CoverageType(models.Model):
    name = models.CharField(max_length=512, unique=True, validators=name_validators, **mandatory)

    def __str__(self):
        return self.name


class ProductType(models.Model):
    name = models.CharField(max_length=512, unique=True, validators=name_validators, **mandatory)
    allowed_coverage_types = models.ManyToManyField(CoverageType, related_name='allowed_product_types', blank=True)

    def __str__(self):
        return self.name


class CollectionType(models.Model):
    name = models.CharField(max_length=512, unique=True, validators=name_validators, **mandatory)
    allowed_coverage_types = models.ManyToManyField(CoverageType, related_name='allowed_collection_types', blank=True)
    allowed_product_types = models.ManyToManyField(ProductType, related_name='allowed_collection_types', blank=True)

    def __str__(self):
        return self.name


class BrowseType(models.Model):
    product_type = models.ForeignKey(ProductType, on_delete=models.CASCADE, related_name="browse_types", **mandatory)
    name = models.CharField(max_length=256, validators=name_validators, blank=True, null=False)

    red_or_grey_expression = models.CharField(max_length=512, validators=[band_expression_validator], **optional)
    green_expression = models.CharField(max_length=512, validators=[band_expression_validator], **optional)
    blue_expression = models.CharField(max_length=512, validators=[band_expression_validator], **optional)
    alpha_expression = models.CharField(max_length=512, validators=[band_expression_validator], **optional)

    red_or_grey_nodata_value = models.FloatField(**optional)
    green_nodata_value = models.FloatField(**optional)
    blue_nodata_value = models.FloatField(**optional)
    alpha_nodata_value = models.FloatField(**optional)

    red_or_grey_range_min = models.FloatField(**optional)
    green_range_min = models.FloatField(**optional)
    blue_range_min = models.FloatField(**optional)
    alpha_range_min = models.FloatField(**optional)

    red_or_grey_range_max = models.FloatField(**optional)
    green_range_max = models.FloatField(**optional)
    blue_range_max = models.FloatField(**optional)
    alpha_range_max = models.FloatField(**optional)

    def __str__(self):
        if self.name:
            return self.name
        return "Default Browse Type for '%s'" % self.product_type

    def clean(self):
        return validate_browse_type(self)

    class Meta:
        unique_together = (
            ('name', 'product_type'),
        )


# ==============================================================================
# Metadata models for each Collection, Product or Coverage
# ==============================================================================


def axis_accessor(pattern, value_map=None):
    def _get(self):
        values = []
        for i in range(1, 5):
            value = getattr(self, pattern % i)
            if value is not None:
                values.append(value_map[value] if value_map else value)
            else:
                break
        return values
    return _get


class Grid(models.Model):
    AXIS_TYPES = [
        (0, 'spatial'),
        (1, 'elevation'),
        (2, 'temporal'),
        (3, 'other'),
    ]

    # allow named grids but also anonymous ones
    name = models.CharField(max_length=256, unique=True, null=True, blank=False, validators=name_validators)

    coordinate_reference_system = models.TextField(**mandatory)

    axis_1_name = models.CharField(max_length=256, **mandatory)
    axis_2_name = models.CharField(max_length=256, **optional)
    axis_3_name = models.CharField(max_length=256, **optional)
    axis_4_name = models.CharField(max_length=256, **optional)

    axis_1_type = models.SmallIntegerField(choices=AXIS_TYPES, **mandatory)
    axis_2_type = models.SmallIntegerField(choices=AXIS_TYPES, **optional)
    axis_3_type = models.SmallIntegerField(choices=AXIS_TYPES, **optional)
    axis_4_type = models.SmallIntegerField(choices=AXIS_TYPES, **optional)

    # using 'char' here, to allow a wide range of datatypes (such as time)
    # when axis_1_offset is null, then this grid is referenceable
    axis_1_offset = models.CharField(max_length=256, **optional)
    axis_2_offset = models.CharField(max_length=256, **optional)
    axis_3_offset = models.CharField(max_length=256, **optional)
    axis_4_offset = models.CharField(max_length=256, **optional)

    resolution = models.PositiveIntegerField(**optional)

    axis_names = property(axis_accessor('axis_%d_name'))
    axis_types = property(axis_accessor('axis_%d_type', dict(AXIS_TYPES)))
    axis_offsets = property(axis_accessor('axis_%d_offset'))

    def __str__(self):
        if self.name:
            return self.name
        elif self.resolution is not None \
                and len(self.coordinate_reference_system) < 15:
            return '%s (%d)' % (
                self.coordinate_reference_system, self.resolution
            )
        return super(Grid, self).__str__()

    def clean(self):
        validate_grid(self)


class GridFixture(models.Model):
    # optional here to allow 'referenceable' coverages
    grid = models.ForeignKey(Grid, **optional_protected)

    axis_1_origin = models.CharField(max_length=256, **optional)
    axis_2_origin = models.CharField(max_length=256, **optional)
    axis_3_origin = models.CharField(max_length=256, **optional)
    axis_4_origin = models.CharField(max_length=256, **optional)

    axis_1_size = models.PositiveIntegerField(**mandatory)
    axis_2_size = models.PositiveIntegerField(**optional)
    axis_3_size = models.PositiveIntegerField(**optional)
    axis_4_size = models.PositiveIntegerField(**optional)

    origin = property(axis_accessor('axis_%d_origin'))
    size = property(axis_accessor('axis_%d_size'))

    class Meta:
        abstract = True


# ==============================================================================
# Actual item models: Collection, Product and Coverage
# ==============================================================================

def eo_object_identifier_validator(value):
    if getattr(settings, 'EOXS_VALIDATE_IDS_NCNAME', True):
        identifier_validators[0](value)


@python_2_unicode_compatible
class EOObject(models.Model):
    """ Base class for Collections, Products and Coverages
    """
    identifier = models.CharField(max_length=256, unique=True, validators=[eo_object_identifier_validator], **mandatory)

    begin_time = models.DateTimeField(**optional)
    end_time = models.DateTimeField(**optional)
    footprint = models.GeometryField(**optional)

    inserted = models.DateTimeField(auto_now_add=True)
    updated = models.DateTimeField(auto_now=True)

    objects = InheritanceManager()

    def __str__(self):
        return self.identifier


class Collection(EOObject):
    collection_type = models.ForeignKey(CollectionType, related_name='collections', **optional_protected)

    grid = models.ForeignKey(Grid, on_delete=models.CASCADE, **optional)


class Mosaic(EOObject, GridFixture):
    coverage_type = models.ForeignKey(CoverageType, related_name='mosaics', **mandatory_protected)

    collections = models.ManyToManyField(Collection, related_name='mosaics', blank=True)


class Product(EOObject):
    product_type = models.ForeignKey(ProductType, related_name='products', **optional_protected)

    collections = models.ManyToManyField(Collection, related_name='products', blank=True)
    package = models.OneToOneField(backends.Storage, **optional_protected)


class Coverage(EOObject, GridFixture):
    coverage_type = models.ForeignKey(CoverageType, related_name='coverages', **optional_protected)

    collections = models.ManyToManyField(Collection, related_name='coverages', blank=True)
    mosaics = models.ManyToManyField(Mosaic, related_name='coverages', blank=True)
    parent_product = models.ForeignKey(Product, on_delete=models.CASCADE, related_name='coverages', **optional)


class ReservedIDManager(models.Manager):
    """ Model manager for `ReservedID` models for easier handling. Returns only
        `QuerySets` that contain valid reservations.
    """
    def get_original_queryset(self):
        return super(ReservedIDManager, self).get_queryset()

    def get_queryset(self):
        Q = models.Q
        return self.get_original_queryset().filter(
            Q(until__isnull=True) | Q(until__gt=now())
        )

    def cleanup_reservations(self):
        Q = models.Q
        self.get_original_queryset().filter(
            Q(until__isnull=False) | Q(until__lte=now())
        ).delete()

    def remove_reservation(self, identifier=None, request_id=None):
        if not identifier and not request_id:
            raise ValueError("Either identifier or request ID required")

        if identifier:
            model = self.get_original_queryset().get(identifier=identifier)
            if request_id and model.request_id != request_id:
                raise ValueError(
                    "Given request ID does not match the reservation."
                )
        else:
            model = self.get_original_queryset().get(request_id=request_id)
            model.delete()


class ReservedID(EOObject):
    """ Model to reserve a specific ID. The field `until` can be used to
        specify the end of the reservation.
        """
    until = models.DateTimeField(**optional)
    request_id = models.CharField(max_length=256, **optional)

    objects = ReservedIDManager()


# ==============================================================================
# DataItems subclasses
# ==============================================================================

class MetaDataItem(backends.DataItem):
    SEMANTIC_CHOICES = [
        (0, 'other'),
        (1, 'description'),
        (2, 'documentation'),
        (3, 'thumbnail'),
    ]

    semantic_names = {
        code: name
        for code, name in SEMANTIC_CHOICES
    }

    semantic_codes = {
        name: code
        for code, name in SEMANTIC_CHOICES
    }

    eo_object = models.ForeignKey(EOObject, on_delete=models.CASCADE, related_name='metadata_items', **mandatory)
    semantic = models.SmallIntegerField(choices=SEMANTIC_CHOICES, **optional)

    class Meta:
        unique_together = [('eo_object', 'semantic')]



class Browse(backends.DataItem):
    product = models.ForeignKey(Product, on_delete=models.CASCADE, related_name='browses', **mandatory)
    browse_type = models.ForeignKey(BrowseType, on_delete=models.CASCADE, **optional)
    style = models.CharField(max_length=256, **optional)

    coordinate_reference_system = models.TextField(**mandatory)
    min_x = models.FloatField(**mandatory)
    min_y = models.FloatField(**mandatory)
    max_x = models.FloatField(**mandatory)
    max_y = models.FloatField(**mandatory)
    width = models.PositiveIntegerField(**mandatory)
    height = models.PositiveIntegerField(**mandatory)

    class Meta:
        unique_together = [('product', 'browse_type', 'style')]


class Mask(backends.DataItem):
    product = models.ForeignKey(Product, on_delete=models.CASCADE, related_name='masks', **mandatory)
    mask_type = models.ForeignKey(MaskType, on_delete=models.CASCADE, **mandatory)

    geometry = models.GeometryField(**optional)


class ProductDataItem(backends.DataItem):
    product = models.ForeignKey(Product, on_delete=models.CASCADE, related_name='product_data_items', **mandatory)


class ArrayDataItem(backends.DataItem):
    BANDS_INTERPRETATION_CHOICES = [
        (0, 'fields'),
        (1, 'dimension')
    ]

    coverage = models.ForeignKey(EOObject, on_delete=models.CASCADE, related_name='arraydata_items', **mandatory)

    field_index = models.PositiveSmallIntegerField(default=0, **mandatory)
    band_count = models.PositiveSmallIntegerField(default=1, **mandatory)

    subdataset_type = models.CharField(max_length=64, **optional)
    subdataset_locator = models.CharField(max_length=1024, **optional)

    bands_interpretation = models.PositiveSmallIntegerField(default=0, choices=BANDS_INTERPRETATION_CHOICES, **mandatory)

    class Meta:
        unique_together = [('coverage', 'field_index')]


# ==============================================================================
# Additional Metadata Models for Collections, Products and Coverages
# ==============================================================================


class CollectionMetadata(models.Model):
    collection = models.OneToOneField(Collection, on_delete=models.CASCADE, related_name='collection_metadata')

    product_type = models.CharField(max_length=256, **optional_indexed)
    doi = models.CharField(max_length=256, **optional_indexed)
    platform = models.CharField(max_length=256, **optional_indexed)
    platform_serial_identifier = models.CharField(max_length=256, **optional_indexed)
    instrument = models.CharField(max_length=256, **optional_indexed)
    sensor_type = models.CharField(max_length=256, **optional_indexed)
    composite_type = models.CharField(max_length=256, **optional_indexed)
    processing_level = models.CharField(max_length=256, **optional_indexed)
    orbit_type = models.CharField(max_length=256, **optional_indexed)
    spectral_range = models.CharField(max_length=256, **optional_indexed)
    wavelength = models.IntegerField(**optional_indexed)
    # hasSecurityConstraints = models.CharField(**optional_indexed)
    # dissemination = models.CharField(**optional_indexed)

    product_metadata_summary = models.TextField(**optional)
    coverage_metadata_summary = models.TextField(**optional)


# ==============================================================================
# "Common value" tables to store string enumerations
# ==============================================================================


@python_2_unicode_compatible
class AbstractCommonValue(models.Model):
    value = models.CharField(max_length=256, db_index=True, unique=True)

    def __str__(self):
        return self.value

    class Meta:
        abstract = True


class OrbitNumber(AbstractCommonValue):
    pass


class Track(AbstractCommonValue):
    pass


class Frame(AbstractCommonValue):
    pass


class SwathIdentifier(AbstractCommonValue):
    pass


class ProductVersion(AbstractCommonValue):
    pass


class ProductQualityDegredationTag(AbstractCommonValue):
    pass


class ProcessorName(AbstractCommonValue):
    pass


class ProcessingCenter(AbstractCommonValue):
    pass


class SensorMode(AbstractCommonValue):
    pass


class ArchivingCenter(AbstractCommonValue):
    pass


class ProcessingMode(AbstractCommonValue):
    pass


class AcquisitionStation(AbstractCommonValue):
    pass


class AcquisitionSubType(AbstractCommonValue):
    pass


class ProductMetadata(models.Model):
    PRODUCTION_STATUS_CHOICES = (
        (0, 'ARCHIVED'),
        (1, 'ACQUIRED'),
        (2, 'CANCELLED')
    )

    ACQUISITION_TYPE_CHOICES = (
        (0, 'NOMINAL'),
        (1, 'CALIBRATION'),
        (2, 'OTHER')
    )

    ORBIT_DIRECTION_CHOICES = (
        (0, 'ASCENDING'),
        (1, 'DESCENDING')
    )

    PRODUCT_QUALITY_STATUS_CHOICES = (
        (0, 'NOMINAL'),
        (1, 'DEGRAGED')
    )

    POLARISATION_MODE_CHOICES = (
        (0, 'single'),
        (1, 'dual'),
        (2, 'twin'),
        (3, 'quad'),
        (4, 'UNDEFINED')
    )

    POLARISATION_CHANNELS_CHOICES = (
        (0, "HV"),
        (1, "HV, VH"),
        (2, "VH"),
        (3, "VV"),
        (4, "HH, VV"),
        (5, "HH, VH"),
        (6, "HH, HV"),
        (7, "VH, VV"),
        (8, "VH, HV"),
        (9, "VV, HV"),
        (10, "VV, VH"),
        (11, "HH"),
        (12, "HH, HV, VH, VV"),
        (13, "UNDEFINED"),
    )

    ANTENNA_LOOK_DIRECTION_CHOICES = (
        (0, 'LEFT'),
        (1, 'RIGHT')
    )

    product = models.OneToOneField(Product, on_delete=models.CASCADE, related_name='product_metadata')

    parent_identifier = models.CharField(max_length=256, **optional_indexed)

    production_status = models.PositiveSmallIntegerField(choices=PRODUCTION_STATUS_CHOICES, **optional_indexed)
    acquisition_type = models.PositiveSmallIntegerField(choices=ACQUISITION_TYPE_CHOICES, **optional_indexed)

    orbit_number = models.ForeignKey(OrbitNumber, **common_value_args)
    orbit_direction = models.PositiveSmallIntegerField(choices=ORBIT_DIRECTION_CHOICES, **optional_indexed)

    track = models.ForeignKey(Track, **common_value_args)
    frame = models.ForeignKey(Frame, **common_value_args)
    swath_identifier = models.ForeignKey(SwathIdentifier, **common_value_args)

    product_version = models.ForeignKey(ProductVersion, **common_value_args)
    product_quality_status = models.PositiveSmallIntegerField(choices=PRODUCT_QUALITY_STATUS_CHOICES, **optional_indexed)
    product_quality_degradation_tag = models.ForeignKey(ProductQualityDegredationTag, **common_value_args)
    processor_name = models.ForeignKey(ProcessorName, **common_value_args)
    processing_center = models.ForeignKey(ProcessingCenter, **common_value_args)
    creation_date = models.DateTimeField(**optional_indexed) # insertion into catalog
    modification_date = models.DateTimeField(**optional_indexed) # last modification in catalog
    processing_date = models.DateTimeField(**optional_indexed)
    sensor_mode = models.ForeignKey(SensorMode, **common_value_args)
    archiving_center = models.ForeignKey(ArchivingCenter, **common_value_args)
    processing_mode = models.ForeignKey(ProcessingMode, **common_value_args)

    # acquisition type metadata
    availability_time = models.DateTimeField(**optional_indexed)
    acquisition_station = models.ForeignKey(AcquisitionStation, **common_value_args)
    acquisition_sub_type = models.ForeignKey(AcquisitionSubType, **common_value_args)
    start_time_from_ascending_node = models.IntegerField(**optional_indexed)
    completion_time_from_ascending_node = models.IntegerField(**optional_indexed)
    illumination_azimuth_angle = models.FloatField(**optional_indexed)
    illumination_zenith_angle = models.FloatField(**optional_indexed)
    illumination_elevation_angle = models.FloatField(**optional_indexed)
    polarisation_mode = models.PositiveSmallIntegerField(choices=POLARISATION_MODE_CHOICES, **optional_indexed)
    polarization_channels = models.PositiveSmallIntegerField(choices=POLARISATION_CHANNELS_CHOICES, **optional_indexed)
    antenna_look_direction = models.PositiveSmallIntegerField(choices=ANTENNA_LOOK_DIRECTION_CHOICES, **optional_indexed)
    minimum_incidence_angle = models.FloatField(**optional_indexed)
    maximum_incidence_angle = models.FloatField(**optional_indexed)
    across_track_incidence_angle = models.FloatField(**optional_indexed)
    along_track_incidence_angle = models.FloatField(**optional_indexed)
    # for SAR acquisitions
    doppler_frequency = models.FloatField(**optional_indexed)
    incidence_angle_variation = models.FloatField(**optional_indexed)
    # for OPT/ALT
    cloud_cover = models.FloatField(**optional_indexed)
    snow_cover = models.FloatField(**optional_indexed)
    lowest_location = models.FloatField(**optional_indexed)
    highest_location = models.FloatField(**optional_indexed)


class CoverageMetadata(models.Model):
    coverage = models.OneToOneField(Coverage, on_delete=models.CASCADE, related_name="coverage_metadata")


# ==============================================================================
# Functions interacting with models. Done here, to keep the model definitions
# as short and concise as possible
# ==============================================================================


class ManagementError(Exception):
    pass


def cast_eo_object(eo_object):
    """ Casts an EOObject to its actual type.
    """
    if isinstance(eo_object, EOObject):
        try:
            return eo_object.collection
        except:
            try:
                return eo_object.mosaic
            except:
                try:
                    return eo_object.product
                except:
                    try:
                        return eo_object.coverage
                    except:
                        pass

    return eo_object


def collection_insert_eo_object(collection, eo_object, use_extent=False):
    """ Inserts an EOObject (either a Product or Coverage) into a collection.
        When an EOObject is passed, it is downcast to its actual type. An error
        is raised when an object of the wrong type is passed.
        The collections footprint and time-stamps are adjusted when necessary.
    """
    collection_type = collection.collection_type
    eo_object = cast_eo_object(eo_object)
    if not isinstance(eo_object, (Product, Coverage)):
        raise ManagementError(
            'Cannot insert object of type %r' % type(eo_object).__name__
        )

    if isinstance(eo_object, Product):
        product_type = eo_object.product_type
        allowed = True
        if collection_type:
            if not product_type:
                allowed = False
            else:
                allowed = collection_type.allowed_product_types.filter(
                    pk=product_type.pk
                ).exists()

        if not allowed:
            raise ManagementError(
                'Cannot insert Product as the product type %r is not allowed in '
                'this collection' % product_type.name
            )

        collection.products.add(eo_object)

    elif isinstance(eo_object, Coverage):
        coverage_type = eo_object.coverage_type
        allowed = True
        if collection_type:
            if not coverage_type:
                allowed = False
            else:
                allowed = collection_type.allowed_coverage_types.filter(
                    pk=coverage_type.pk
                ).exists()

        if not allowed:
            raise ManagementError(
                'Cannot insert Coverage as the coverage type %r is not allowed '
                'in this collection' % coverage_type.name
            )

        if collection.grid and collection.grid != eo_object.grid:
            raise ManagementError(
                'Cannot insert Coverage as the coverage grid is not '
                'compatible with this collection'
            )

        collection.coverages.add(eo_object)

    if eo_object.footprint:
        footprint = eo_object.footprint
        if use_extent:
            footprint = Polygon.from_bbox(footprint.extent)

        if collection.footprint:
            collection.footprint = collection.footprint.union(footprint)
            if use_extent:
                collection.footprint = Polygon.from_bbox(
                    collection.footprint.extent
                )
        else:
            collection.footprint = eo_object.footprint

    if eo_object.begin_time:
        collection.begin_time = (
            eo_object.begin_time if not collection.begin_time
            else min(eo_object.begin_time, collection.begin_time)
        )

    if eo_object.end_time:
        collection.end_time = (
            eo_object.end_time if not collection.end_time
            else max(eo_object.end_time, collection.end_time)
        )

    collection.full_clean()
    collection.save()


def collection_exclude_eo_object(collection, eo_object, use_extent=False):
    """ Exclude an EOObject (either Product or Coverage) from the collection.
    """
    eo_object = cast_eo_object(eo_object)

    if not isinstance(eo_object, (Product, Coverage)):
        raise ManagementError(
            'Cannot exclude object of type %r' % type(eo_object).__name__
        )

    if isinstance(eo_object, Product):
        collection.products.remove(eo_object)

    elif isinstance(eo_object, Coverage):
        collection.coverages.remove(eo_object)

    collection_collect_metadata(
        collection,
        eo_object.footprint is not None,
        eo_object.begin_time and eo_object.begin_time == collection.begin_time,
        eo_object.end_time and eo_object.end_time == collection.end_time,
        use_extent=use_extent
    )
    collection.full_clean()
    collection.save()


def collection_collect_metadata(collection, collect_footprint=True,
                                collect_begin_time=True, collect_end_time=True,
                                product_summary=False, coverage_summary=False,
                                use_extent=False):
    """ Collect metadata
    """

    if collect_footprint or collect_begin_time or collect_end_time:
        aggregates = {}

        if collect_footprint:
            if use_extent:
                aggregates["extent"] = Extent("footprint")
            else:
                aggregates["footprint"] = Union("footprint")
        if collect_begin_time:
            aggregates["begin_time"] = Min("begin_time")
        if collect_end_time:
            aggregates["end_time"] = Max("end_time")

        values = EOObject.objects.filter(
            Q(coverage__collections=collection) |
            Q(product__collections=collection)
        ).aggregate(**aggregates)

        if collect_footprint:
            if use_extent:
                collection.footprint = Polygon.from_bbox(
                    values["extent"]
                )
            else:
                collection.footprint = values["footprint"]
        if collect_begin_time:
            collection.begin_time = values["begin_time"]
        if collect_end_time:
            collection.end_time = values["end_time"]

    if product_summary or coverage_summary:
        collection_metadata, _ = CollectionMetadata.objects.get_or_create(
            collection=collection
        )

        if product_summary:
            collection_metadata.product_metadata_summary = json.dumps(
                _collection_metadata(
                    collection, ProductMetadata, 'product'
                ), indent=4, sort_keys=True
            )

        if coverage_summary:
            collection_metadata.coverage_metadata_summary = json.dumps(
                _collection_metadata(
                    collection, CoverageMetadata, 'coverage'
                ), indent=4, sort_keys=True
            )

        collection_metadata.save()


def _collection_metadata(collection, metadata_model, path):
    summary_metadata = {}
    fields = metadata_model._meta.get_fields()

    def is_common_value(field):
        if isinstance(field, models.ForeignKey):
            return issubclass(field.related_model, AbstractCommonValue)
        return False

    # "Value fields": float, ints, dates, etc; displaying a single value
    value_fields = [
        field for field in fields
        if isinstance(field, (
            models.FloatField, models.IntegerField, models.DateTimeField
        )) and not field.choices
    ]

    # choice fields
    choice_fields = [
        field for field in fields if field.choices
    ]

    # "common value" fields
    common_value_fields = [
        field for field in fields
        if is_common_value(field)
    ]

    base_query = metadata_model.objects.filter(
        **{"%s__collections__in" % path: [collection]}
    )

    # get a list of all related common values
    for field in common_value_fields:
        summary_metadata[field.name] = list(
            field.related_model.objects.filter(
                **{"metadatas__%s__collections" % path: collection}
            ).values_list('value', flat=True).distinct()
        )

    # get a list of all related choice fields
    for field in choice_fields:
        summary_metadata[field.name] = [
            dict(field.choices)[raw_value]
            for raw_value in base_query.filter(
                **{"%s__isnull" % field.name: False}
            ).values_list(
                field.name, flat=True
            ).distinct()
        ]

    # get min/max
    aggregates = {}
    for field in value_fields:
        aggregates.update({
            "%s_min" % field.name: Min(field.name),
            "%s_max" % field.name: Max(field.name),
        })
    values = base_query.aggregate(**aggregates)

    for field in value_fields:
        min_ = values["%s_min" % field.name]
        max_ = values["%s_max" % field.name]

        if isinstance(min_, datetime):
            min_ = isoformat(min_)
        if isinstance(max_, datetime):
            max_ = isoformat(max_)

        summary_metadata[field.name] = {
            "min": min_,
            "max": max_,
        }

    return summary_metadata


def mosaic_insert_coverage(mosaic, coverage):
    """ Insert a coverage into a mosaic.
    """

    mosaic = cast_eo_object(mosaic)
    coverage = cast_eo_object(coverage)

    assert isinstance(mosaic, Mosaic)
    assert isinstance(coverage, Coverage)

    grid = mosaic.grid

    if mosaic.coverage_type != coverage.coverage_type:
        raise ManagementError(
            'Cannot insert Coverage %s as its coverage type does not match '
            'the Mosaics coverage type.' % coverage
        )
    elif grid and grid != coverage.grid:
        raise ManagementError(
            'Cannot insert Coverage %s as its grid does not match '
            'the Mosaics grid.' % coverage
        )

    mosaic.coverages.add(coverage)

    mosaic_recalc_metadata(mosaic)
    mosaic.full_clean()
    mosaic.save()


def mosaic_exclude_coverage(mosaic, coverage):
    """ Exclude a coverage from a mosaic.
    """
    mosaic = cast_eo_object(mosaic)
    coverage = cast_eo_object(coverage)

    assert isinstance(mosaic, Mosaic)
    assert isinstance(coverage, Coverage)

    if not mosaic.coverages.filter(identifier=coverage.identifier).exists():
        raise ManagementError(
            'Cannot exclude Coverage %s as it is not contained in the mosaic.'
            % coverage
        )

    mosaic.coverages.remove(coverage)

    mosaic_recalc_metadata(mosaic)
    mosaic.full_clean()
    mosaic.save()



def mosaic_recalc_metadata(mosaic):
    """ Recalculates axis origins and time/footprint metadata
        for the given mosaic model. Does not save the model.
    """
    values = mosaic.coverages.aggregate(begin_time=Min('begin_time'),end_time=Max('end_time'),footprint=Union('footprint'),)
    mosaic.begin_time = values['begin_time']
    mosaic.end_time = values['end_time']
    mosaic.footprint = values['footprint']

    grid = mosaic.grid
    if grid:
        # compute new origins and size

        casts = {}
        annotations = {}
        aggregates = {}
        for i in range(1, 5):
            if getattr(grid, 'axis_%d_type' % i) is None:
                break

            offset = float(getattr(grid, 'axis_%d_offset' % i))

            casts['axis_%d_origin_float' % i] = Cast(
                'axis_%d_origin' % i, models.FloatField()
            )
            annotations['axis_%d_top' % i] = ExpressionWrapper(
                F('axis_%d_origin_float' % i) + F('axis_%d_size' % i) * offset,
                output_field=models.FloatField(),
            )

            if offset < 0:
                aggregates['axis_%d_min' % i] = Max('axis_%d_origin_float' % i)
                aggregates['axis_%d_max' % i] = Min('axis_%d_top' % i)
            else:
                aggregates['axis_%d_min' % i] = Min('axis_%d_origin_float' % i)
                aggregates['axis_%d_max' % i] = Max('axis_%d_top' % i)

        # create a query to get the min/max for each axis
        values = mosaic.coverages.annotate(
            **casts
        ).annotate(
            **annotations
        ).aggregate(
            **aggregates
        )

        for i in range(1, 5):
            if 'axis_%d_min' % i not in values:
                break

            offset = float(getattr(grid, 'axis_%d_offset' % i))

            if values['axis_%d_min' % i] is not None:
                setattr(
                    mosaic, 'axis_%d_size' % i,
                    (values['axis_%d_max' % i] - values['axis_%d_min' % i]) / offset
                )
                setattr(
                    mosaic, 'axis_%d_origin' % i,
                    values['axis_%d_min' % i]
                )
            else:
                setattr(mosaic, 'axis_%d_size' % i, 0)
                setattr(mosaic, 'axis_%d_origin' % i, None)


def product_add_coverage(product, coverage):
    """ Add a Coverage to a product.
        When an EOObject is passed, it is downcast to its actual type. An error
        is raised when an object of the wrong type is passed.
        The collections footprint and time-stamps are adjusted when necessary.
    """
    coverage = cast_eo_object(coverage)
    if not isinstance(coverage, Coverage):
        raise ManagementError(
            'Cannot insert object of type %r' % type(coverage).__name__
        )

    product_type = product.product_type
    coverage_type = coverage.coverage_type

    allowed = True
    if product_type:
        if not coverage_type:
            allowed = False
        else:
            allowed = product_type.allowed_coverage_types.filter(
                pk=coverage_type.pk
            ).exists()

    if not allowed:
        raise ManagementError(
            'Cannot insert Coverage as the coverage type %r is not allowed '
            'in this product' % (coverage_type.name if coverage_type else 'None')
        )

    product.coverages.add(coverage)

# ==============================================================================
# Validators
# ==============================================================================


def validate_grid(grid):
    """ Validation function for grids.
    """

    higher_dim = False
    for i in range(4, 0, -1):
        axis_type = getattr(grid, 'axis_%d_type' % i, None)
        axis_name = getattr(grid, 'axis_%d_name' % i, None)
        axis_offset = getattr(grid, 'axis_%d_offset' % i, None)

        attrs = (axis_type, axis_name, axis_offset)

        has_dim = any(attrs)

        # check that when this axis is not set, no higher axis is set
        if not has_dim and higher_dim:
            raise ValidationError(
                'Axis %d not set, but higher axis %d is set.' % (i, higher_dim)
            )

        # check that all of 'name', 'type', and 'offset' is set
        if has_dim and not all(attrs):
            raise ValidationError(
                "For each axis, 'name', 'type', and 'offset' must be set."
            )

        higher_dim = i if has_dim else False


def validate_browse_type(browse_type):
    """ Validate the expressions of the browse type to only reference fields
        available for that browse type.
    """
    expressions = [
        browse_type.red_or_grey_expression,
        browse_type.green_expression,
        browse_type.blue_expression,
        browse_type.alpha_expression,
    ]

    fields = set()
    for expression in expressions:
        try:
            fields |= set(extract_fields(browse_type.red_or_grey_expression))
        except BandExpressionError:
            pass

    all_fields = set(
        FieldType.objects.filter(
            coverage_type__allowed_product_types=browse_type.product_type,
        ).values_list('identifier', flat=True)
    )

    missing_fields = fields - all_fields
    if missing_fields:
        raise ValidationError(
            "Expressions are referencing unknow field%s: %s. Available field%s: "
            "%s." % (
                "s" if len(missing_fields) > 1 else "",
                ", ".join(("'%s'" % field) for field in missing_fields),
                "s" if len(all_fields) > 1 else "",
                ", ".join(("'%s'" % field) for field in all_fields),
            )
        )

# ==============================================================================
# Utilities
# ==============================================================================


def product_get_metadata(product):
    try:
        product_metadata = product.product_metadata
    except ProductMetadata.DoesNotExist:
        return []

    def get_value(product_metadata, field):
        raw_value = getattr(product_metadata, field.name)
        if isinstance(field, models.ForeignKey):
            return raw_value.value
        elif field.choices:
            return dict(field.choices)[raw_value]
        return raw_value

    return [
        (field.name, get_value(product_metadata, field))
        for field in ProductMetadata._meta.fields
        if field.name not in ('id', 'product') and
        getattr(product_metadata, field.name, None) is not None
    ]



#-------------------------------------------------------------------------------
#
# Project: EOxServer <http://eoxserver.org>
# Authors: Fabian Schindler <fabian.schindler@eox.at>
#
#-------------------------------------------------------------------------------
# Copyright (C) 2013 EOX IT Services GmbH
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies of this Software or works derived from this Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.
#-------------------------------------------------------------------------------


from eoxserver.core.decoders import kvp, xml, upper, typelist
from eoxserver.core.util.xmltools import NameSpace, NameSpaceMap
from eoxserver.services.ows.version import parse_version_string


ns_xlink = NameSpace("http://www.w3.org/1999/xlink", "xlink")
ns_ows10 = NameSpace("http://www.opengis.net/ows/1.0", "ows10")
ns_ows11 = NameSpace("http://www.opengis.net/ows/1.1", "ows11")
ns_ows20 = NameSpace("http://www.opengis.net/ows/2.0", "ows20")

nsmap = NameSpaceMap(ns_ows10, ns_ows11, ns_ows20)


def get_decoder(request):
    """ Convenience function to return the right OWS Common request deocder for
        the given `django.http.HttpRequest`.
    """
    if request.method == "GET":
        return OWSCommonKVPDecoder(request.GET)
    elif request.method == "POST":
        # TODO: this may also be in a different format.
        return OWSCommonXMLDecoder(request.body)


class OWSCommonKVPDecoder(kvp.Decoder):
    service         = kvp.Parameter("service", type=upper, num="?")
    version         = kvp.Parameter("version", type=parse_version_string, num="?")
    request         = kvp.Parameter("request", type=upper)
    acceptversions  = kvp.Parameter(type=typelist(parse_version_string, ","), num="?")


class OWSCommonXMLDecoder(xml.Decoder):
    service         = xml.Parameter("@service", type=upper, num="?")
    version         = xml.Parameter("@version", type=parse_version_string, num="?")
    request         = xml.Parameter("local-name()", type=upper)
    acceptversions  = xml.Parameter(
        "ows10:AcceptVersions/ows10:Version/text() "
        "| ows11:AcceptVersions/ows11:Version/text() "
        "| ows20:AcceptVersions/ows20:Version/text()",
        type=parse_version_string, num="*"
    )

    namespaces = nsmap



# ------------------------------------------------------------------------------
#
# Project: EOxServer <http://eoxserver.org>
# Authors: Fabian Schindler <fabian.schindler@eox.at>
#
# ------------------------------------------------------------------------------
# Copyright (C) 2017 EOX IT Services GmbH
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies of this Software or works derived from this Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.
# ------------------------------------------------------------------------------

import os
from os.path import join, isdir, isfile
import zipfile

from django.contrib.gis.geos import Polygon, MultiPolygon

from eoxserver.core.util.xmltools import parse as parse_xml
from eoxserver.core.util.timetools import parse_iso8601


nsmap = {
    'safe': 'http://www.esa.int/safe/sentinel-1.0',
    'gml': 'http://www.opengis.net/gml'
}


class S1ProductFormatReader(object):
    def test_path(self, path):
        try:
            manifest = self.open_manifest(path)
            if not manifest:
                return False

            root = manifest.getroot()
            return root.xpath(
                'metadataSection/metadataObject[@ID="platform"]/metadataWrap/'
                'xmlData/safe:platform/safe:familyName/text()',
                namespaces=nsmap
            )[0] == "SENTINEL-1"
        except (IOError, RuntimeError, IndexError):
            return False

    def read_path(self, path):
        values = {}
        root = self.open_manifest(path).getroot()

        period_elem = root.xpath(
            'metadataSection/metadataObject[@ID="acquisitionPeriod"]/'
            'metadataWrap/xmlData/safe:acquisitionPeriod',
            namespaces=nsmap
        )[0]

        values['begin_time'] = parse_iso8601(
            period_elem.findtext('safe:startTime', namespaces=nsmap)
        )
        values['end_time'] = parse_iso8601(
            period_elem.findtext('safe:stopTime', namespaces=nsmap)
        )

        coordinates_elems = root.xpath(
            'metadataSection/metadataObject[@ID="measurementFrameSet"]/'
            'metadataWrap/xmlData/safe:frameSet/safe:frame/safe:footPrint/'
            'gml:coordinates',
            namespaces=nsmap
        )

        values['footprint'] = MultiPolygon([
            self.parse_coordinates(coordinates_elem.text)
            for coordinates_elem in coordinates_elems
        ]).wkt

        # values['identifier'] =

        # values['browses'] = [
        #     (None, tci_path(granule))
        # ]

        # TODO: extended metadata

        # values['parent_identifier']
        # values['production_status']
        # values['acquisition_type']
        # values['orbit_number'] =
        # values['orbit_direction'] =
        # values['track']
        # values['frame']
        # values['swath_identifier'] =
        # values['product_version'] =
        # values['product_quality_status']
        # values['product_quality_degradation_tag']
        # values['processor_name']
        # values['processing_center']
        # values['creation_date']
        # values['modification_date']
        # values['processing_date'] =
        # values['sensor_mode']
        # values['archiving_center'] =
        # values['processing_mode']

        # values['availability_time'] =
        # values['acquisition_station']
        # values['acquisition_sub_type']
        # values['start_time_from_ascending_node']
        # values['completion_time_from_ascending_node']
        # values['illumination_azimuth_angle'] =
        # values['illumination_zenith_angle'] =
        # values['illumination_elevation_angle']
        # values['polarisation_mode']
        # values['polarization_channels']
        # values['antenna_look_direction']
        # values['minimum_incidence_angle']
        # values['maximum_incidence_angle']

        # values['doppler_frequency']
        # values['incidence_angle_variation']

        # values['cloud_cover'] =
        # values['snow_cover']
        # values['lowest_location']
        # values['highest_location']

        return values

    def open_manifest(self, path):
        """ Tries to open the manifest of a given sentinel 1 SAFE product.
            Supported are:
                - directories containing a Sentinel-1 product
                - zip files containing a sentinel product
                - a direct path reference to a safe file
        """
        if isdir(path):
            manifest_path = join(path, 'manifest.safe')
            if not isfile(manifest_path):
                try:
                    manifest_path = join(
                        path,
                        get_immediate_subdirectories(path)[0],
                        'manifest.safe'
                    )
                except IndexError:
                    raise IOError(
                        "Could not locate 'manifest.safe' in %r" % path
                    )

            with open(manifest_path) as f:
                return parse_xml(f)
        elif zipfile.is_zipfile(path):
            with zipfile.ZipFile(path) as zp_f:
                names = [
                    name for name in zp_f.namelist()
                    if name.endswith('manifest.safe')
                ]

                try:
                    return parse_xml(zp_f.open(names[0]))
                except IndexError:
                    raise IOError("Could not find 'manifest.safe' in %r" % path)

        elif isfile(path):
            with open(path) as f:
                return parse_xml(f)

        raise IOError('Could not open manifest for path %r' % path)

    def parse_coordinates(self, coords, swap=True):
        points = [
            tuple(float(v) for v in coord.split(','))
            for coord in coords.split()
        ]

        if swap:
            points[:] = [
                (p[1], p[0])
                for p in points
            ]

        points.append(points[0])
        return Polygon(points)


def get_immediate_subdirectories(a_dir):
    return [
        name
        for name in os.listdir(a_dir)
        if isdir(join(a_dir, name))
    ]



# -*- coding: utf-8 -*-
# Generated by Django 1.11.3 on 2017-08-28 10:02
from __future__ import unicode_literals

from django.db import migrations, models
import django.db.models.deletion


class Migration(migrations.Migration):

    initial = True

    dependencies = [
        ('coverages', '0001_initial'),
    ]

    operations = [
        migrations.CreateModel(
            name='Extra',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('display_name', models.CharField(blank=True, max_length=64, null=True)),
                ('info', models.TextField(blank=True, null=True)),
                ('color', models.CharField(blank=True, max_length=64, null=True)),
                ('default_visible', models.BooleanField(default=False)),
                ('eo_object', models.OneToOneField(on_delete=django.db.models.deletion.CASCADE, related_name='webclient_extra', to='coverages.EOObject')),
            ],
        ),
    ]



#-------------------------------------------------------------------------------
#
# Project: EOxServer <http://eoxserver.org>
# Authors: Stephan Krause <stephan.krause@eox.at>
#          Stephan Meissl <stephan.meissl@eox.at>
#
#-------------------------------------------------------------------------------
# Copyright (C) 2011 EOX IT Services GmbH
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies of this Software or works derived from this Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.
#-------------------------------------------------------------------------------

"""
This module contains basic classes and functions for the security layer (which
is integrated in the service layer for now).
"""

import logging

from django.http import HttpResponse

from eoxserver.core import Component, ExtensionPoint, env
from eoxserver.core.config import get_eoxserver_config
from eoxserver.core.decoders import config
from eoxserver.services.auth.exceptions import AuthorisationException
from eoxserver.services.auth.interfaces import PolicyDecisionPointInterface


logger = logging.getLogger(__name__)

#-------------------------------------------------------------------------------
# Config reader
#-------------------------------------------------------------------------------

class AuthConfigReader(config.Reader):
    section = "services.auth.base"


    attribute_mapping = config.Option()
    authz_service = config.Option()
    serviceID = config.Option()
    allowLocal = config.Option(type=bool)
    pdp_type = config.Option()


#-------------------------------------------------------------------------------
# PDP Base Class
#-------------------------------------------------------------------------------

class BasePDP(object):
    """
    This is the base class for PDP implementations. It provides a skeleton for
    authorization request handling.
    """

    def authorize(self, request):
        """
        This method handles authorization requests according to the
        requirements given in the :class:`PolicyDecisionPointInterface`
        declaration.

        Internally, it invokes the :meth:`_decide` method that implements the
        actual authorization decision logic.
        """

        reader = AuthConfigReader(get_eoxserver_config())

        # This code segment allows local clients bypassing the
        # Authorisation process.
        if reader.allowLocal:
            remoteAddress = request.META['REMOTE_ADDR']
            # Check all possibilities, also IPv6
            if remoteAddress in ('127.0.0.1', 'localhost', '::1'):
                return True

        authorized, message = self._decide(request)
        if authorized:
            return True
        else:
            raise AuthorisationException(message)


    def _decide(self, request):

        # This method shall implement the actual authorization decision
        # logic. It gets an :class:`~.OWSRequest` object as input and shall
        # return a tuple of ``(authorized, message)`` where ``authorized`` is
        # a boolean flag and ``message`` is a string containing an error
        # message in case authorization is not granted.
        #
        # This method must be overridden by concrete implementations.

        return (True, "")

#-------------------------------------------------------------------------------
# utility functions
#-------------------------------------------------------------------------------

class PDPComponent(Component):
    pdps = ExtensionPoint(PolicyDecisionPointInterface)

    def get_pdp(self, pdp_type):
        for pdp in self.pdps:
            if pdp.pdp_type == pdp_type:
                return pdp
        return None


def getPDP():
    reader = AuthConfigReader(get_eoxserver_config())
    if not reader.pdp_type or reader.pdp_type == "none":
        logger.debug("Authorization deactivated.")
        return None
    else:
        return PDPComponent(env).get_pdp(reader.pdp_type)



# ------------------------------------------------------------------------------
#
# Project: EOxServer <http://eoxserver.org>
# Authors: Fabian Schindler <fabian.schindler@eox.at>
#
# ------------------------------------------------------------------------------
# Copyright (C) 2017 EOX IT Services GmbH
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies of this Software or works derived from this Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.
# ------------------------------------------------------------------------------

from django.core.management.base import CommandError, BaseCommand
from django.db import transaction

from eoxserver.resources.coverages import models
from eoxserver.resources.coverages.management.commands import (
    CommandOutputMixIn, SubParserMixIn
)


class Command(CommandOutputMixIn, SubParserMixIn, BaseCommand):
    """ Command to manage collection types. This command uses sub-commands for the
        specific tasks: create, delete
    """
    def add_arguments(self, parser):
        create_parser = self.add_subparser(parser, 'create')
        delete_parser = self.add_subparser(parser, 'delete')
        list_parser = self.add_subparser(parser, 'list')

        # identifier is a common argument
        for parser in [create_parser, delete_parser]:
            parser.add_argument(
                'name', nargs=1, help='The collection type name. Mandatory.'
            )

        create_parser.add_argument(
            '--coverage-type', '-c', action='append', default=[],
            dest='allowed_coverage_type_names',
            help=(
                'Specify a coverage type that is allowed in collections of this '
                'type.'
            )
        )
        create_parser.add_argument(
            '--product-type', '-p', action='append', default=[],
            dest='allowed_product_type_names',
            help=(
                'Specify a product type that is allowed in collections of this '
                'type.'
            )
        )

        delete_parser.add_argument(
            '--force', '-f', action='store_true', default=False,
            help='Also remove all collections associated with that type.'
        )

        list_parser.add_argument(
            '--no-detail', action="store_false", default=True, dest='detail',
            help="Disable the printing of details of the collection type."
        )

    @transaction.atomic
    def handle(self, subcommand, *args, **kwargs):
        """ Dispatch sub-commands: create, delete, insert and exclude.
        """
        if subcommand == "create":
            self.handle_create(kwargs.pop('name')[0], *args, **kwargs)
        elif subcommand == "delete":
            self.handle_delete(kwargs.pop('name')[0], *args, **kwargs)
        elif subcommand == "list":
            self.handle_list(*args, **kwargs)

    def handle_create(self, name, allowed_coverage_type_names,
                      allowed_product_type_names, **kwargs):
        """ Handle the creation of a new collection type.
        """

        collection_type = models.CollectionType.objects.create(name=name)

        for allowed_coverage_type_name in allowed_coverage_type_names:
            try:
                collection_type.allowed_coverage_types.add(
                    models.CoverageType.objects.get(
                        name=allowed_coverage_type_name
                    )
                )
            except models.CoverageType.DoesNotExist:
                raise CommandError(
                    'Coverage type %r does not exist.' %
                    allowed_coverage_type_name
                )

        for allowed_product_type_name in allowed_product_type_names:
            try:
                collection_type.allowed_product_types.add(
                    models.ProductType.objects.get(
                        name=allowed_product_type_name
                    )
                )
            except models.ProductType.DoesNotExist:
                raise CommandError(
                    'Product type %r does not exist.' %
                    allowed_product_type_name
                )

        print('Successfully created collection type %r' % name)

    def handle_delete(self, name, force, **kwargs):
        """ Handle the deletion of a collection type
        """
        collection_type = models.CollectionType.objects.get(name=name)
        collection_type.delete()
        # TODO: force

        print('Successfully deleted collection type %r' % name)

    def handle_list(self, detail, *args, **kwargs):
        """ Handle the listing of product types
        """
        for collection_type in models.CollectionType.objects.all():
            print(collection_type.name)
            # if detail:
            #     for coverage_type in collection_type.allowed_coverage_types.all():
            #         print("\t%s" % coverage_type.name)



#-------------------------------------------------------------------------------
#
# Project: EOxServer <http://eoxserver.org>
# Authors: Fabian Schindler <fabian.schindler@eox.at>
#
#-------------------------------------------------------------------------------
# Copyright (C) 2011 EOX IT Services GmbH
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies of this Software or works derived from this Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.
#-------------------------------------------------------------------------------

from django.conf import settings
import traceback
from itertools import chain

from lxml import etree
from lxml.builder import ElementMaker

from eoxserver.core.util.xmltools import XMLEncoder, NameSpace, NameSpaceMap
from eoxserver.services.ows.dispatch import filter_handlers


ns_xlink = NameSpace("http://www.w3.org/1999/xlink", "xlink")
ns_ows = NameSpace("http://www.opengis.net/ows/2.0", "ows", "http://schemas.opengis.net/ows/2.0/owsAll.xsd")
ns_xml = NameSpace("http://www.w3.org/XML/1998/namespace", "xml")

nsmap = NameSpaceMap(ns_ows)
OWS = ElementMaker(namespace=ns_ows.uri, nsmap=nsmap)


class OWS20Encoder(XMLEncoder):
    def get_conf(self):
        raise NotImplementedError

    def get_http_service_url(self, request):
        conf = self.get_conf()
        if conf.http_service_url:
            return conf.http_service_url
        from eoxserver.services.urls import get_http_service_url
        return get_http_service_url(request)

    def encode_reference(self, node_name, href, reftype="simple"):

        attributes = {ns_xlink("href"): href}
        if reftype:
            attributes[ns_xlink("type")] = reftype

        return OWS(node_name, **attributes)

    def encode_service_identification(self, service, conf, profiles):
        # get a list of versions in descending order from all active
        # GetCapabilities handlers.
        handlers = filter_handlers(
            service=service, request="GetCapabilities"
        )
        versions = sorted(
            set(chain(*[handler.versions for handler in handlers])),
            reverse=True
        )

        elem = OWS("ServiceIdentification",
            OWS("Title", conf.title),
            OWS("Abstract", conf.abstract),
            OWS("Keywords", *[
                OWS("Keyword", keyword) for keyword in conf.keywords
            ]),
            OWS("ServiceType", "OGC WCS", codeSpace="OGC")
        )

        elem.extend(
            OWS("ServiceTypeVersion", version) for version in versions
        )

        elem.extend(
            OWS("Profile", "http://www.opengis.net/%s" % profile)
            for profile in profiles
        )

        elem.extend((
            OWS("Fees", conf.fees),
            OWS("AccessConstraints", conf.access_constraints)
        ))
        return elem

    def encode_service_provider(self, conf):
        return OWS("ServiceProvider",
            OWS("ProviderName", conf.provider_name),
            self.encode_reference("ProviderSite", conf.provider_site),
            OWS("ServiceContact",
                OWS("IndividualName", conf.individual_name),
                OWS("PositionName", conf.position_name),
                OWS("ContactInfo",
                    OWS("Phone",
                        OWS("Voice", conf.phone_voice),
                        OWS("Facsimile", conf.phone_facsimile)
                    ),
                    OWS("Address",
                        OWS("DeliveryPoint", conf.delivery_point),
                        OWS("City", conf.city),
                        OWS("AdministrativeArea", conf.administrative_area),
                        OWS("PostalCode", conf.postal_code),
                        OWS("Country", conf.country),
                        OWS(
                            "ElectronicMailAddress",
                            conf.electronic_mail_address
                        )
                    ),
                    self.encode_reference(
                        "OnlineResource", conf.onlineresource
                    ),
                    OWS("HoursOfService", conf.hours_of_service),
                    OWS("ContactInstructions", conf.contact_instructions)
                ),
                OWS("Role", conf.role)
            )
        )

    def encode_operations_metadata(self, request, service, versions):
        get_handlers = filter_handlers(
            service=service, versions=versions, method="GET"
        )
        post_handlers = filter_handlers(
            service=service, versions=versions, method="POST"
        )
        all_handlers = sorted(
            set(get_handlers + post_handlers),
            key=lambda h: (getattr(h, "index", 10000), h.request)
        )

        http_service_url = self.get_http_service_url(request)

        operations = []
        for handler in all_handlers:
            methods = []
            if handler in get_handlers:
                methods.append(
                    self.encode_reference("Get", http_service_url)
                )
            if handler in post_handlers:
                post = self.encode_reference("Post", http_service_url)
                post.append(
                    OWS("Constraint",
                        OWS("AllowedValues",
                            OWS("Value", "XML")
                        ), name="PostEncoding"
                    )
                )
                methods.append(post)

            operations.append(
                OWS("Operation",
                    OWS("DCP",
                        OWS("HTTP", *methods)
                    ),
                    # apply default values as constraints
                    *([
                        OWS("Constraint",
                            OWS("NoValues"),
                            OWS("DefaultValue", str(default)),
                            name=name
                        )
                        for name, default
                        in getattr(handler(), "constraints", {}).items()
                    ] + [
                        OWS("Parameter",
                            OWS("AnyValue")
                            if allowed_values is None else
                            OWS("AllowedValues", *[
                                OWS("Value", value)
                                for value in allowed_values
                            ]),
                            name=name
                        )
                        for name, allowed_values
                        in getattr(handler(), "additional_parameters", {}).items()
                    ]),

                    name=handler.request
                )
            )

        return OWS("OperationsMetadata", *operations)


class OWS20ExceptionXMLEncoder(XMLEncoder):
    def encode_exception(self, message, version, code, locator=None):
        exception_attributes = {
            "exceptionCode": str(code)
        }

        if locator:
            exception_attributes["locator"] = str(locator)

        exception_text = (OWS("ExceptionText", message),) if message else ()

        report = OWS("ExceptionReport",
            OWS("Exception", *exception_text, **exception_attributes),
            version=version, **{ns_xml("lang"): "en"}
        )

        try:
            if getattr(settings, 'DEBUG', False):
                report.append(etree.Comment(traceback.format_exc()))
        except:
            pass

        return report

    def get_schema_locations(self):
        return nsmap.schema_locations



#-------------------------------------------------------------------------------
#
#  WPS Complex Data codecs (encoders/decoders)
#
# Project: EOxServer <http://eoxserver.org>
# Authors: Martin Paces <martin.paces@eox.at>
#
#-------------------------------------------------------------------------------
# Copyright (C) 2014 EOX IT Services GmbH
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies of this Software or works derived from this Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.
#-------------------------------------------------------------------------------

from base64 import (
    standard_b64encode, standard_b64decode, urlsafe_b64encode, urlsafe_b64decode,
)


class Codec(object):
    """ Base complex data codec."""
    encoding = None

    @staticmethod
    def encode(file_in, **opt):
        """ Encoding generator."""
        raise NotImplementedError

    @staticmethod
    def decode(file_in, **opt):
        """ Encoding generator."""
        raise NotImplementedError


class CodecBase64(Codec):
    """ Base64 codec """
    encoding = 'base64'

    @staticmethod
    def encode(file_in, urlsafe=False, **opt):
        """ Encoding generator."""
        b64encode = urlsafe_b64encode if urlsafe else standard_b64encode
        dlm = b""

        for data in iter(lambda: file_in.read(57), b''):
            yield dlm
            yield b64encode(data)
            dlm = b"\r\n"

    @staticmethod
    def decode(file_in, urlsafe=False, **opt):
        """ Decoding generator."""
        b64decode = urlsafe_b64decode if urlsafe else standard_b64decode
        for data in file_in:
            yield b64decode(data)


class CodecRaw(Codec):
    """ Data encoder class."""
    encoding = None

    @staticmethod
    def encode(file_in, **opt):
        """ Encoding generator."""
        for data in iter(lambda: file_in.read(65536), ''):
            yield data

    @staticmethod
    def decode(file_in, **opt):
        """ Decoding generator."""
        for data in iter(lambda: file_in.read(65536), ''):
            yield data



#-------------------------------------------------------------------------------
#
# Project: EOxServer <http://eoxserver.org>
# Authors: Fabian Schindler <fabian.schindler@eox.at>
#
#-------------------------------------------------------------------------------
# Copyright (C) 2014 EOX IT Services GmbH
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies of this Software or works derived from this Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.
#-------------------------------------------------------------------------------


class RegistratorInterface(object):
    """ Interface for components that allow the registration files as datasets.
    """

    def register(self, items, overrides=None, cache=None):
        """ Register a dataset as the given ``dataset_type`` and the provided
        ``items``. All required metadata is retrieved from the
        ``items``, overrides can be specified by supplying an ``overrides``
        :class:`dict`.

        :param items: an iterable that yields four-tuples:
                      (storage or package or ``None``, location, semantic,
                      format).
        :param overrides: a :class:`dict` containing any metadata value that
                          cannot be retrieved from the supplied ``data_items``,
                          or shall override any supplied value
        :param cache: an instance of :class:`CacheContext
                      <eoxserver.backends.cache.CacheContext>` for cached file
                      access during registration
        :returns: the registered dataset. the actual type depends on the passed
                  metadata
        """
        pass



#-------------------------------------------------------------------------------
#
# WPS 1.0 parameters' XML encoders
#
# Project: EOxServer <http://eoxserver.org>
# Authors: Fabian Schindler <fabian.schindler@eox.at>
#          Martin Paces <martin.paces@eox.at>
#
#-------------------------------------------------------------------------------
# Copyright (C) 2013 EOX IT Services GmbH
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies of this Software or works derived from this Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.
#-------------------------------------------------------------------------------
# pylint: disable=bad-continuation
from django.utils.six import itervalues

from eoxserver.services.ows.wps.parameters import (
    LiteralData, ComplexData, BoundingBoxData,
    AllowedAny, AllowedEnum, AllowedRange, AllowedRangeCollection,
    AllowedByReference, RequestParameter
)

from eoxserver.services.ows.wps.v10.util import (
    OWS, WPS, NIL, ns_ows,
)


def encode_input_descr(prm):
    """ Encode process description Input element."""
    # Request input parameters are not visible from the process description.
    if isinstance(prm, RequestParameter):
        return None
    elem = NIL("Input", *_encode_param_common(prm))
    elem.attrib["minOccurs"] = ("1", "0")[bool(prm.is_optional)]
    elem.attrib["maxOccurs"] = "1"
    if isinstance(prm, LiteralData):
        elem.append(_encode_literal(prm, True))
    elif isinstance(prm, ComplexData):
        elem.append(_encode_complex(prm, True))
    elif isinstance(prm, BoundingBoxData):
        elem.append(_encode_bbox(prm, True))
    return elem


def encode_output_descr(prm):
    """ Encode process description Output element."""
    elem = NIL("Output", *_encode_param_common(prm))
    if isinstance(prm, LiteralData):
        elem.append(_encode_literal(prm, False))
    elif isinstance(prm, ComplexData):
        elem.append(_encode_complex(prm, False))
    elif isinstance(prm, BoundingBoxData):
        elem.append(_encode_bbox(prm, False))
    return elem


def encode_input_exec(prm):
    """ Encode common part of the execute response Input (data) element. """
    return WPS("Input", *_encode_param_common(prm, False))


def encode_output_exec(prm):
    """ Encode common part of the execute response Output (data) element. """
    return WPS("Output", *_encode_param_common(prm))


def encode_output_def(outdef):
    """ Encode execute response Output (definition) element. """
    attrib = {}
    if outdef.uom is not None:
        attrib['uom'] = outdef.uom
    if outdef.crs is not None:
        attrib['crs'] = outdef.crs
    if outdef.mime_type is not None:
        attrib['mimeType'] = outdef.mime_type
    if outdef.encoding is not None:
        attrib['encoding'] = outdef.encoding
    if outdef.schema is not None:
        attrib['schema'] = outdef.schema
    if outdef.as_reference is not None:
        attrib['asReference'] = 'true' if outdef.as_reference else 'false'
    return WPS("Output", *_encode_param_common(outdef, False), **attrib)


def _encode_param_common(prm, title_required=True):
    """ Encode the Identifier, Title and Abstract sub-elements common to all
    input and output parameters.
    """
    elist = [OWS("Identifier", prm.identifier)]
    if prm.title or title_required:
        elist.append(OWS("Title", prm.title or prm.identifier))
    if prm.abstract:
        elist.append(OWS("Abstract", prm.abstract))
    return elist


def _encode_literal(prm, is_input):
    """ Encode process description LiteralData (parameter definition) element.
    """
    dtype = prm.dtype
    elem = NIL("LiteralData" if is_input else "LiteralOutput")

    elem.append(OWS("DataType", dtype.name, **{
        ns_ows("reference"): "http://www.w3.org/TR/xmlschema-2/#%s"%dtype.name,
    }))

    if prm.uoms:
        elem.append(NIL("UOMs",
            NIL("Default", OWS("UOM", prm.uoms[0])),
            NIL("Supported", *[OWS("UOM", u) for u in prm.uoms])
        ))

    if is_input:
        elem.append(_encode_allowed_value(prm.allowed_values))

        if prm.default is not None:
            elem.append(NIL("DefaultValue", str(prm.default)))

    return elem


def _encode_allowed_value(avobj):
    """ Encode process description LiteralData allowed values definition.
    Based on the type of the allowed value definition either AnyValue,
    ValuesReference, or AllowedValues element is returned.
    """
    enum, ranges, elist = None, [], []

    if isinstance(avobj, AllowedAny):
        return OWS("AnyValue")
    elif isinstance(avobj, AllowedByReference):
        return WPS("ValuesReference", **{
            ns_ows("reference"): avobj.url,
            "valuesForm": avobj.url,
        })
    elif isinstance(avobj, AllowedEnum):
        enum = avobj
    elif isinstance(avobj, AllowedRange):
        ranges = [avobj]
    elif isinstance(avobj, AllowedRangeCollection):
        enum, ranges = avobj.enum, avobj.ranges
    else:
        raise TypeError("Invalid allowed value object! OBJ=%r"%avobj)

    dtype = avobj.dtype
    ddtype = dtype.get_diff_dtype()

    if enum is not None:
        elist.extend(OWS("Value", dtype.encode(v)) for v in enum.values)
    for range_ in ranges:
        attr, elms = {}, []
        if range_.closure != 'closed':
            attr = {ns_ows("rangeClosure"): range_.closure}
        if range_.minval is not None:
            elms.append(OWS("MinimumValue", dtype.encode(range_.minval)))
        if range_.maxval is not None:
            elms.append(OWS("MaximumValue", dtype.encode(range_.maxval)))
        if range_.spacing is not None:
            elms.append(OWS("Spacing", ddtype.encode(range_.spacing)))
        elist.append(OWS("Range", *elms, **attr))

    return OWS("AllowedValues", *elist)


def _encode_complex(prm, is_input):
    """ Encode process description ComplexData (parameter definition) element.
    """
    return NIL("ComplexData" if is_input else "ComplexOutput",
        NIL("Default", _encode_format(prm.default_format)),
        NIL("Supported", *[_encode_format(f) for f in itervalues(prm.formats)])
    )


def _encode_format(frmt):
    """ Encode process description ComplexData format definition.
    The function returns Format element.
    """
    elem = NIL("Format", NIL("MimeType", frmt.mime_type))
    if frmt.encoding is not None:
        elem.append(NIL("Encoding", frmt.encoding))
    if frmt.schema is not None:
        elem.append(NIL("Schema", frmt.schema))
    return elem


def _encode_bbox(prm, is_input):
    """ Encode process description BoundingBoxData (parameter definition)
    element.
    """
    return NIL("BoundingBoxData" if is_input else "BoundingBoxOutput",
        NIL("Default", NIL("CRS", prm.encode_crs(prm.default_crs))),
        NIL("Supported", *[NIL("CRS", prm.encode_crs(crs)) for crs in prm.crss])
    )



#-------------------------------------------------------------------------------
#
# Project: EOxServer <http://eoxserver.org>
# Authors: Fabian Schindler <fabian.schindler@eox.at>
#
#-------------------------------------------------------------------------------
# Copyright (C) 2016 EOX IT Services GmbH
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies of this Software or works derived from this Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.
#-------------------------------------------------------------------------------

# from https://gist.github.com/kulturlupenguen/69aec1259131b5619fb7

from django.template import Library
from django.template.defaulttags import URLNode, url

register = Library()


class AbsoluteURL(str):
    pass


class AbsoluteURLNode(URLNode):
    def render(self, context):
        asvar, self.asvar = self.asvar, None
        path = super(AbsoluteURLNode, self).render(context)
        request_obj = context['request']
        abs_url = AbsoluteURL(request_obj.build_absolute_uri(path))

        if not asvar:
            return str(abs_url)
        else:
            if path == request_obj.path:
                abs_url.active = 'active'
            else:
                abs_url.active = ''
            context[asvar] = abs_url
            return ''


@register.tag
def absurl(parser, token):
    node = url(parser, token)
    return AbsoluteURLNode(
        view_name=node.view_name,
        args=node.args,
        kwargs=node.kwargs,
        asvar=node.asvar
    )



# ------------------------------------------------------------------------------
#
# Project: EOxServer <http://eoxserver.org>
# Authors: Fabian Schindler <fabian.schindler@eox.at>
#
# ------------------------------------------------------------------------------
# Copyright (C) 2019 EOX IT Services GmbH
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies of this Software or works derived from this Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.
# ------------------------------------------------------------------------------

from eoxserver.backends import models
from eoxserver.backends.storages import (
    get_handler_by_test, get_handler_class_by_name
)


def resolve_storage(storage_paths, save=True):
    parent = None
    for locator in storage_paths:
        # try to get a storage by name, if it exists:
        try:
            parent = models.Storage.objects.get(
                name=locator, parent=parent
            )
        except models.Storage.DoesNotExist:
            handler = get_handler_by_test(locator)
            if handler:
                parent = models.Storage(
                    name=None,
                    url=locator, storage_type=handler.name,
                    parent=parent,
                )
            else:
                name, _, url = locator.partition(':')
                handler_cls = get_handler_class_by_name(name)
                if handler_cls:
                    parent = models.Storage(
                        name=None,
                        url=url, storage_type=name,
                        parent=parent,
                    )
                else:
                    raise Exception(
                        'No storage handler found for locator %r' % locator
                    )
            if save:
                parent.full_clean()
                parent.save()

    return parent


def resolve_storage_and_path(storage_paths, save=True):
    last = storage_paths[-1]
    name, _, _ = last.partition(':')
    handler_cls = get_handler_class_by_name(name)
    if handler_cls:
        return resolve_storage(storage_paths, save), None
    else:
        return resolve_storage(storage_paths[:-1], save), last



#-------------------------------------------------------------------------------
#
# Project: EOxServer <http://eoxserver.org>
# Authors: Fabian Schindler <fabian.schindler@eox.at>
#
#-------------------------------------------------------------------------------
# Copyright (C) 2014 EOX IT Services GmbH
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies of this Software or works derived from this Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.
#-------------------------------------------------------------------------------

try:
    from django.conf.urls import url as re_path
except ImportError:
    from django.urls import re_path
try:
    from django.core.urlresolvers import reverse
except ImportError:
    from django.urls import reverse

from eoxserver.services import views

urlpatterns = [
    re_path(r'^$', views.ows, name='ows')
]


def get_http_service_url(request=None):
    """ Returns the URL the OWS view is available under. If a
        :class:`django.http.HttpRequest` is passed, an absolute URL is
        constructed with the request information.
    """

    http_service_url = reverse("ows")
    if request:
        http_service_url = request.build_absolute_uri(http_service_url)
    return http_service_url



# -*- coding: utf-8 -*-
# Generated by Django 1.11.23 on 2020-02-06 15:13
from __future__ import unicode_literals

import django.core.validators
from django.db import migrations, models
import re


class Migration(migrations.Migration):

    dependencies = [
        ('coverages', '0006_masktype_validity'),
    ]

    operations = [
        migrations.AlterField(
            model_name='arraydataitem',
            name='bands_interpretation',
            field=models.PositiveSmallIntegerField(choices=[(0, 'fields'), (1, 'dimension')], default=0),
        ),
        migrations.AlterField(
            model_name='browsetype',
            name='name',
            field=models.CharField(blank=True, max_length=256, validators=[django.core.validators.RegexValidator(re.compile('^[a-zA-z_][a-zA-Z0-9_]*$', 32), message='This field must contain a valid Name.')]),
        ),
        migrations.AlterField(
            model_name='collectiontype',
            name='name',
            field=models.CharField(max_length=512, unique=True, validators=[django.core.validators.RegexValidator(re.compile('^[a-zA-z_][a-zA-Z0-9_]*$', 32), message='This field must contain a valid Name.')]),
        ),
        migrations.AlterField(
            model_name='coveragetype',
            name='name',
            field=models.CharField(max_length=512, unique=True, validators=[django.core.validators.RegexValidator(re.compile('^[a-zA-z_][a-zA-Z0-9_]*$', 32), message='This field must contain a valid Name.')]),
        ),
        migrations.AlterField(
            model_name='fieldtype',
            name='identifier',
            field=models.CharField(max_length=512, validators=[django.core.validators.RegexValidator(re.compile('^[a-zA-z_][a-zA-Z0-9_.-]*$', 32), message='This field must contain a valid NCName.')]),
        ),
        migrations.AlterField(
            model_name='grid',
            name='axis_1_type',
            field=models.SmallIntegerField(choices=[(0, 'spatial'), (1, 'elevation'), (2, 'temporal'), (3, 'other')]),
        ),
        migrations.AlterField(
            model_name='grid',
            name='axis_2_type',
            field=models.SmallIntegerField(blank=True, choices=[(0, 'spatial'), (1, 'elevation'), (2, 'temporal'), (3, 'other')], null=True),
        ),
        migrations.AlterField(
            model_name='grid',
            name='axis_3_type',
            field=models.SmallIntegerField(blank=True, choices=[(0, 'spatial'), (1, 'elevation'), (2, 'temporal'), (3, 'other')], null=True),
        ),
        migrations.AlterField(
            model_name='grid',
            name='axis_4_type',
            field=models.SmallIntegerField(blank=True, choices=[(0, 'spatial'), (1, 'elevation'), (2, 'temporal'), (3, 'other')], null=True),
        ),
        migrations.AlterField(
            model_name='grid',
            name='name',
            field=models.CharField(max_length=256, null=True, unique=True, validators=[django.core.validators.RegexValidator(re.compile('^[a-zA-z_][a-zA-Z0-9_]*$', 32), message='This field must contain a valid Name.')]),
        ),
        migrations.AlterField(
            model_name='masktype',
            name='name',
            field=models.CharField(max_length=512, validators=[django.core.validators.RegexValidator(re.compile('^[a-zA-z_][a-zA-Z0-9_]*$', 32), message='This field must contain a valid Name.')]),
        ),
        migrations.AlterField(
            model_name='metadataitem',
            name='semantic',
            field=models.SmallIntegerField(blank=True, choices=[(0, 'other'), (1, 'description'), (2, 'documentation'), (3, 'thumbnail')], null=True),
        ),
        migrations.AlterField(
            model_name='nilvalue',
            name='reason',
            field=models.CharField(choices=[('http://www.opengis.net/def/nil/OGC/0/inapplicable', 'Inapplicable (There is no value)'), ('http://www.opengis.net/def/nil/OGC/0/missing', 'Missing'), ('http://www.opengis.net/def/nil/OGC/0/template', 'Template (The value will be available later)'), ('http://www.opengis.net/def/nil/OGC/0/unknown', 'Unknown'), ('http://www.opengis.net/def/nil/OGC/0/withheld', 'Withheld (The value is not divulged)'), ('http://www.opengis.net/def/nil/OGC/0/AboveDetectionRange', 'Above detection range'), ('http://www.opengis.net/def/nil/OGC/0/BelowDetectionRange', 'Below detection range')], max_length=512),
        ),
        migrations.AlterField(
            model_name='productmetadata',
            name='acquisition_type',
            field=models.PositiveSmallIntegerField(blank=True, choices=[(0, 'NOMINAL'), (1, 'CALIBRATION'), (2, 'OTHER')], db_index=True, null=True),
        ),
        migrations.AlterField(
            model_name='productmetadata',
            name='antenna_look_direction',
            field=models.PositiveSmallIntegerField(blank=True, choices=[(0, 'LEFT'), (1, 'RIGHT')], db_index=True, null=True),
        ),
        migrations.AlterField(
            model_name='productmetadata',
            name='orbit_direction',
            field=models.PositiveSmallIntegerField(blank=True, choices=[(0, 'ASCENDING'), (1, 'DESCENDING')], db_index=True, null=True),
        ),
        migrations.AlterField(
            model_name='productmetadata',
            name='polarisation_mode',
            field=models.PositiveSmallIntegerField(blank=True, choices=[(0, 'single'), (1, 'dual'), (2, 'twin'), (3, 'quad'), (4, 'UNDEFINED')], db_index=True, null=True),
        ),
        migrations.AlterField(
            model_name='productmetadata',
            name='polarization_channels',
            field=models.PositiveSmallIntegerField(blank=True, choices=[(0, 'HV'), (1, 'HV, VH'), (2, 'VH'), (3, 'VV'), (4, 'HH, VV'), (5, 'HH, VH'), (6, 'HH, HV'), (7, 'VH, VV'), (8, 'VH, HV'), (9, 'VV, HV'), (10, 'VV, VH'), (11, 'HH'), (12, 'HH, HV, VH, VV'), (13, 'UNDEFINED')], db_index=True, null=True),
        ),
        migrations.AlterField(
            model_name='productmetadata',
            name='product_quality_status',
            field=models.PositiveSmallIntegerField(blank=True, choices=[(0, 'NOMINAL'), (1, 'DEGRAGED')], db_index=True, null=True),
        ),
        migrations.AlterField(
            model_name='productmetadata',
            name='production_status',
            field=models.PositiveSmallIntegerField(blank=True, choices=[(0, 'ARCHIVED'), (1, 'ACQUIRED'), (2, 'CANCELLED')], db_index=True, null=True),
        ),
        migrations.AlterField(
            model_name='producttype',
            name='name',
            field=models.CharField(max_length=512, unique=True, validators=[django.core.validators.RegexValidator(re.compile('^[a-zA-z_][a-zA-Z0-9_]*$', 32), message='This field must contain a valid Name.')]),
        ),
    ]



#-------------------------------------------------------------------------------
#
# Project: EOxServer <http://eoxserver.org>
# Authors: Fabian Schindler <fabian.schindler@eox.at>
#
#-------------------------------------------------------------------------------
# Copyright (C) 2013 EOX IT Services GmbH
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies of this Software or works derived from this Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.
#-------------------------------------------------------------------------------

from eoxserver.core import Component, implements
from eoxserver.services.ows.interfaces import ExceptionHandlerInterface
from eoxserver.services.ows.common.v11.encoders import OWS11ExceptionXMLEncoder


class WPS10ExceptionHandler(Component):
    """ WPS 1.0 exception handler. """
    implements(ExceptionHandlerInterface)

    service = "WPS"
    versions = ("1.0.0", "1.0")
    request = None

    def handle_exception(self, request, exception):
        """ Handle exception. """
        # pylint: disable=unused-argument, no-self-use

        code = getattr(exception, "code", None)
        locator = getattr(exception, "locator", None)
        http_status_code = getattr(exception, "http_status_code", 400)

        if not code:
            code = "NoApplicableCode"
            locator = type(exception).__name__

        encoder = OWS11ExceptionXMLEncoder()

        return (
            encoder.serialize(
                encoder.encode_exception(
                    str(exception), "1.1.0", code, locator
                )
            ),
            encoder.content_type, http_status_code
        )
